# -*- coding: utf-8 -*-

# AUTHORS:
#   - Pierre Lairez (2019): initial implementation
#   - Eric Pichon-Pharabod (2022): adding the coordinates functionality
#   - Eric Pichon-Pharabod (2023): adding the `shift` option


from sage.arith.misc import random_prime
from sage.combinat.integer_vector import IntegerVectors
from sage.geometry.voronoi_diagram import VoronoiDiagram
from sage.graphs.graph import Graph
from sage.matrix.constructor import Matrix
from sage.misc.cachefunc import cached_method
from sage.modules.free_module import FreeModule
from sage.modules.free_module_element import vector
from sage.rings.complex_double import CDF
from sage.rings.complex_interval_field import ComplexIntervalField
from sage.rings.finite_rings.finite_field_constructor import FiniteField
from sage.rings.polynomial.polynomial_ring import *
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.real_double import RDF
from sage.rings.integer_ring import ZZ
from sage.rings.real_mpfi import RealIntervalField
from sage.rings.imaginary_unit import I

from ore_algebra import OreAlgebra

import logging
import random
import signal

from . import interpolation
from . import cohomology
from . import config
from ..exceptions import FailFast

logger = logging.getLogger(__name__)

class Family(object):

    def __init__(self, pol, denom=1, path=None, discoverbasis=False, shift=0):
        """pol is an element of a ring of the form K[x1,...,xn][t]

        This class aims at computing in H^n( P^n - V(pol(t)) ).
        Classes in H^n can be differentiated wrt t, this is the Gauss-Manin connection.
        If m is a monomial (not depending on t), then d/dt [m] = [ -pol.derivative()*m ]
        """

        assert isinstance(pol.parent(), sage.rings.polynomial.polynomial_ring.PolynomialRing_integral_domain)
        assert pol.base_ring().base_ring().is_field()
        self.base_field = pol.base_ring().base_ring()

        coeffs = pol.coefficients()
        assert all(c.is_homogeneous() for c in coeffs)
        assert pol(123).is_homogeneous()

        self.pol = pol
        self.shift = shift
        self.upolring = self.pol.parent().change_ring(self.base_field)
        self.denom = self.upolring(denom)

        if path is None:
            self._path = [ZZ(0),ZZ(1)]
            self._explicit_path = False
        else:
            self._path = path
            self._explicit_path = True

        self.coho1 = self.cohomologyAt(self._path[-1])

        self.discoverbasis = discoverbasis
        if not discoverbasis:
            self.basis = self.coho1.basis()
            self.endpoint = self._path[-1]

        self.dopring = OreAlgebra(self.upolring, 'D' + str(self.upolring.gen()))


    @cached_method
    def cohomologyAt(self, t):
        return cohomology.Cohomology(self.pol(self.pol.base_ring()(t))/self.denom(self.pol.base_ring()(t)), shift=self.shift)

    def modulo(self, prime):
        return Family(FiniteField(prime).one() * self.pol, denom=FiniteField(prime).one()*self.denom, shift=self.shift)

    def _gaussmanin(self, pt):
        logger.debug("Evaluating cohomology at a point %s" % str(pt))
        try:
            co = self.cohomologyAt(pt)
        except cohomology.NotSmoothError:
            raise ZeroDivisionError  # FunctionReconstruction only handles this exception

        der = (self.pol.derivative()(pt)*self.denom(pt) - self.denom.derivative()(pt)*self.pol(pt))/self.denom(pt)**2
        redmul = Matrix([co.coordinates(-b*der) for b in self.basis])
        redb = Matrix([co.coordinates(b) for b in self.basis])

        # Matrices are row-based.
        return redmul*redb.inverse()

    @cached_method
    def gaussmanin(self):
        """Return a pair (mat, denom), where mat is a matrix with polynomial
        coefficients, denom a polynomial and mat/denom the matrix of the Gauss-Manin
        connection in the basis self.basis.

        That is denom*derivative(self.basis) = mat * self.basis. In particular,
        if an element of the homology is given as the scalar product u*basis =
        sum_i u[i]*basis[i], then the derivative of u*basis is (u' +
        u*mat/denom)*basis, where u' is the coefficient-wise derivative.

        """
        

        logger.info("Computing Gauss-Manin connection")
        fr = interpolation.FunctionReconstruction(self.upolring, self._gaussmanin)
        return fr.recons(denomapart=True)

	
    def _coordinates(self, ws, pt):
        logger.debug("Evaluating cohomology at a point %s" % str(pt))
        try:
            co = self.cohomologyAt(pt)
        except cohomology.NotSmoothError:
            raise ZeroDivisionError  # FunctionReconstruction only handles this exception

        redb = Matrix([co.coordinates(b) for b in self.basis])
        coords = Matrix([co.coordinates(w(pt)) for w in ws])

        return coords*redb.inverse()

    def coordinates(self, ws):
        """Returns a list of vectors of coordinates in the basis self.basis, for w in ws

        """
        

        logger.info("Computing coordinates")
        fr = interpolation.FunctionReconstruction(self.upolring, lambda pt: self._coordinates(ws, pt))
        return fr.recons(denomapart=True)
    
    @cached_method
    def picard_fuchs_equation(self, vec=None, form=None):
        """vec is a constant-coefficient vector representing an element omega of
        H^n(P^n - V(pol)) in the basis self.basis.

        """
        #logger.info("Computing a cyclic space.")

        if config.fail_fast:
            signal.alarm(config.time_to_compute_picard_fuchs_equations)

        if form is not None:
            vec = self.coho1.coordinates(form)

        vec = vec.change_ring(self.upolring)
        mat, denom = self.gaussmanin()
        dim = len(self.basis)

        logger.info("Computing Picard-Fuchs equation for %s." % str(vec(1) * vector(self.basis)))

        while True:
            cyclicspace = vec.row()

            rpoint = self.base_field(ZZ.random_element(10000, 100000))
            cyclicspace_at_r = cyclicspace(rpoint)
            var = self.upolring.gen()

            k = 0
            while cyclicspace_at_r.rank() == cyclicspace_at_r.nrows():
                logger.info("Looking for equation of order %d." % (k+1))
                k = k+1
                vec = denom*vec.derivative(var) + vec*mat - (k-1)*denom.derivative(var)*vec
                cyclicspace = denom*cyclicspace
                cyclicspace = cyclicspace.stack(vec)
                cyclicspace_at_r = cyclicspace_at_r.stack(vec(rpoint))

                # The k-th row of cyclicspace is denom^(n-1) * d^k/dt^k [vec], where n is
                # the number of columns. cyclicspace_at_r is the evaluation (up to a row
                # scaling) of cyclicspace at a random point. We test if the rank of cyclicspace is
                # defficient by computing the rank of cyclicspace_at_r. This may fail but we don't
                # care.

            try:
                logger.info("Computing kernel.")
                kernel = cyclicspace.transpose().change_ring(self.upolring.fraction_field()).right_kernel_matrix()
                deq = kernel.row(0)
            except IndexError:
                logger.warn("The matrix equation has no solution, we retry.")
                logger.warn("If this loops for ever, this is an error.")
                continue
            break

        if config.fail_fast:
            signal.alarm(0)

        deq = deq.denominator() * deq

        # This is the differential equation satisfied by the basis element b.
        deq = self.dopring(deq.list())

        logger.info("Found an equation of order %d." % deq.order())
        return deq


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from sage.misc.flatten import flatten
from sage.graphs.graph import Graph
from sage.groups.free_group import FreeGroup

from sage.rings.complex_mpfr import ComplexField

from .util import Util

import logging

logger = logging.getLogger(__name__)


CC = ComplexField(53)

class Translator(object):
    def __init__(self, A, B):
        """A and B are fundamental group classes, such that the points of A are a subset of the points of B
        The basepoint can be distinct. The goal of this class is to find paths of B that represent the same homotopy classes as the paths of A.
        If the basepoints are distinct, then the homotopy class of the loop around A.basepoint is also computed as a path of B.
        """
        
        self.A = A
        self.B = B


    @property
    def AtoB(self):
        if not hasattr(self, "_AtoB"):
            self._AtoB = [Util.select_closest_index(self.B.qpoints, p) for p in self.A.qpoints]
        return self._AtoB

    
    @property
    def edges_tree(self):
        """This tree will serve as the alphabet for the fundamental group"""
        if not hasattr(self, "_edges_tree"):
            minimal_cover_tree = Graph(len(self.A.qpoints)) 
            edges = flatten([[(i,j) for i in range(j)] for j in range(len(self.A.qpoints))], max_level=1) 
            edges.sort(key=(lambda e: Util.simple_rational(abs(CC(self.A.qpoints[e[0]]-self.A.qpoints[e[1]])), 10e-10))) # we sort edges by length
            for e in edges:
                if len(minimal_cover_tree.shortest_path(e[0], e[1]))==0:
                    minimal_cover_tree.add_edge(e)
            self._edges_tree = [list(e)[:2] for e in minimal_cover_tree.edges()]
        return self._edges_tree
    
    @property
    def alphabet(self):
        if not hasattr(self, "_alphabet"):
            self._alphabet = FreeGroup(len(self.edges_tree))
        return self._alphabet
    
    @property
    def letters(self):
        return self.alphabet.gens()
    
    @property
    def Bduality(self):
        if not hasattr(self, "_Bduality"):
            edgesB = []
            for e,d in self.B.duality:
                if d not in edgesB and list(reversed(d)) not in edgesB:
                    edgesB += [d]

            delaunay = Graph()
            for e in edgesB:
                dist = Util.simple_rational(abs(CC(self.B.qpoints[e[0]]-self.B.qpoints[e[1]])), 10**-10)
                delaunay.add_edge(e + [dist])

            edgesA = [[self.AtoB[i] for i in e] for e in self.edges_tree]
            paths = []
            for e in edgesA:
                paths += [delaunay.shortest_path(e[0], e[1], by_weight=True)] 

            Bduality = []
            for dA, path in zip(self.edges_tree, paths):
                for i in range(len(path)-1):
                    e = path[i:i+2]
                    for e2,d2 in self.B.duality:
                        if d2 == e:
                            Bduality += [[e2, dA]]
                            break
            self._Bduality = Bduality
        return self._Bduality

    def wordA(self, path):
        """Given a path of A, return its word in terms of the tree"""
        return self.word(path, self.A.duality, self.edges_tree, self.alphabet)
    def wordB(self, path):
        """Given a path of B, return its word in terms of the tree"""
        return self.word(path, self.Bduality, self.edges_tree, self.alphabet)
    
    @classmethod
    def word(self, path, duality, edges, alphabet=None):
        dual_temp = [[e,d] for e,d in duality if d in edges]
        dual_temp.sort(key=lambda e: edges.index(e[1]))
        edges_temp = [e for e,_ in dual_temp]
        if alphabet==None:
            alphabet = FreeGroup(len(edges))
        w = alphabet(1)
        letters = alphabet.gens()
        for i in range(len(path)-1):
            e = path[i:i+2]
            if e in edges_temp:
                dual = dual_temp[edges_temp.index(e)][1]
                w = letters[edges.index(dual)]**-1 * w
            e = list(reversed(e))
            if e in edges_temp:
                dual = dual_temp[edges_temp.index(e)][1]
                w = letters[edges.index(dual)] * w
        return w
    
    @property
    def thin_gens(self):
        if not hasattr(self, "_thin_gens"):
            self._thin_gens = [self.B.pointed_loops[i-1] for i in self.AtoB[1:]]
        return self._thin_gens
    
    @property
    def fat_gens(self):
        if not hasattr(self, "_fat_gens"):
            points = [i-1 for i in self.AtoB[1:]]
            paths = [[]]
            first=True
            for i, path in enumerate(self.B.pointed_loops):
                if i in points:
                    if first:
                        first = False
                    else:
                        paths += [[]]
                if i!=self.AtoB[0]-1:
                    paths[-1] += path
            self._fat_gens = [Util.simplify_path(path) for path in paths]
        return self._fat_gens
    
    @property
    def lift(self):
        if not hasattr(self, "_lift"):
            words = [self.wordB(path) for path in self.fat_gens]
            proj = self.alphabet.hom(words)
            self._lift = Util.invert_morphism(proj)
        return self._lift
    
    def specialize_path(self, path):
        """Take a path of A and yield a path B with the same homotopy class"""
        w = self.lift(self.wordA(path))
        # conj = Util.middle(w)
        
        path = []
        for l, p in list(reversed(w.syllables())):
            assert p in [-1,1]
            i = self.letters.index(l)
            if p==1:
                path = path + self.fat_gens[i]
            if p==-1:
                path = path + list(reversed(self.fat_gens[i]))
        return Util.simplify_path(path)
        


# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import time

from sage.all import (
    block_matrix,
    ComplexBallField,
    diagonal_matrix,
    Frac,
    matrix,
    MatrixSpace,
    RBF,
    ZZ
)

from ore_algebra import OreAlgebra
from ore_algebra.analytic.accuracy import PrecisionError
from ore_algebra.analytic.bounds import BoundPrecisionError
from ore_algebra.analytic.context import Context, dctx
from ore_algebra.analytic.dac_sum import HighestSolMapper_dac
from ore_algebra.analytic.differential_operator import DifferentialOperator
from ore_algebra.analytic.path import EvaluationPoint_step, Path
from ore_algebra.analytic.utilities import prec_from_eps
from ore_algebra.tools import clear_denominators


logger = logging.getLogger(__name__)

# 2024-12-17 En théorie on pourrait généraliser tout ça aux points singuliers
# réguliers : pour calculer ∫RY sur un lacet basé en z0 ≈ 0, on intègre
# formellement terme à terme RY où R est la même matrice de fractions
# rationnelles que d'habitude et Y est une matrice de séries logarithmiques
# tronquées (qui arrivent par blocs).

class StreamOperation:
    r"""
    Some infrastructure for operations on streams of coefficients

    The aim is to be able to experiment with several ways of implementing our
    integrator.
    """

    def __init__(self, parent, arity, block_size, shift=0):
        # parent = polynomial ring in which we are working
        # (not strictly necessary: for more flexibility, we could start with
        # ints and rely on coercion)
        self.parent = parent
        # block_size = how many coeffs process_block() consumes at once
        # XXX Pas sûr qu'on veuille vraiment fixer ça à l'initialisation, la
        # plupart des opérations sont en réalité plus souples. Et il faut
        # peut-être découpler la taille du bloc qu'on extrait du buffer de
        # celle qu'on produit en sortie.
        assert block_size > 0
        self.block_size = block_size
        self.shift = shift
        self.input = [parent.zero()]*arity
        self.input_len = [0]*arity
        self.input_done = [False]*arity
        self.pos = 0
        # XXX return instead of writing to self.output?
        self.output = parent.zero()
        self.subscribers = []

    def __repr__(self):
        return type(self).__name__

    def subscribe(self, series, operand=0):
        self.subscribers.append((series, operand))

    def push_coefficients(self, coeff, push_len, operand):
        assert coeff.degree() < push_len
        logger.debug("%s received %s coefficients (block size=%s)", self, coeff.degree(), self.block_size)
        self.input[operand] += (coeff << self.input_len[operand])
        self.input_len[operand] += push_len
        # XXX on jette le dernier bloc incomplet reçu, et de même récursivement
        # pour les opérations en aval, c'est un peu merdique
        while all(l >= self.block_size for l in self.input_len):
            self._do_process_block()

    def _do_process_block(self):
        logger.debug("%s processing block at pos %s", self, self.pos)
        block = [input[:self.block_size] for input in self.input]
        self.process_block(*block)
        for i in range(len(self.input)):
            self.input[i] >>= self.block_size
            self.input_len[i] -= self.block_size
            self.input_len[i] = max(0, self.input_len[i])
        push_len = self.block_size
        if self.pos == 0:
            push_len += self.shift
        assert push_len >= 0
        for ser, op in self.subscribers:
            ser.push_coefficients(self.output, push_len, op)
        self.output = self.parent.zero()
        self.pos += self.block_size

    def process_block(self):
        raise NotImplementedError

    def close_input(self, operand=None):
        if operand is not None:
            self.input_done[operand] = True
        if all(self.input_done):
            logger.debug("%s done", self)
            # WARNING for ops with multiple inputs, this consumes all available
            # coefficients, even if some inputs have more data than others
            while any(self.input_len):
                self._do_process_block()
            for ser, op in self.subscribers:
                ser.close_input(op)


class Source(StreamOperation):

    def __init__(self, parent, block_size):
        super().__init__(parent, arity=0, block_size=block_size)

    def push_coefficients(self, coeff, push_len, operand):
        raise RuntimeError

    def generate(self, block):
        assert block.degree() < self.block_size
        for (ser, op) in self.subscribers:
            ser.push_coefficients(block, self.block_size, op)

    def process_block(self):  # XXX
        pass


class Capture(StreamOperation):  # for testing

    def __init__(self, series):
        super().__init__(series.parent, arity=1, block_size=series.block_size)
        series.subscribe(self)
        self.value = series.parent.zero()

    def process_block(self, block):
        self.value += block << self.pos


class Diff(StreamOperation):

    def __init__(self, series):
        super().__init__(series.parent, arity=1, block_size=series.block_size,
                         shift=-1)
        series.subscribe(self)

    def process_block(self, block):
        if self.pos == 0:
            self.output = block.derivative()
        else:  # XXX
            self.output = (block << self.pos).derivative() >> (self.pos - 1)


class Int(StreamOperation):

    def __init__(self, series):
        super().__init__(series.parent, arity=1, block_size=series.block_size,
                         shift=+1)
        series.subscribe(self)

    def process_block(self, block):
        if self.pos == 0:
            self.output = block.integral()
        else:  # XXX
            self.output = (block << self.pos).integral() >> (self.pos + 1)


class MulByPoly(StreamOperation):

    def __init__(self, poly, series):
        assert poly.parent() is series.parent
        super().__init__(poly.parent(), arity=1, block_size=max(poly.degree(), 0) + 1)
        self.poly = poly
        self.state = series.parent.zero()
        series.subscribe(self)

    def __repr__(self):
        return f"{type(self).__name__}(deg={self.poly.degree()})"

    def process_block(self, block):
        # XXX Le gros du temps passe dans la ligne suivante. Normal, mais
        # peut-on faire mieux ? Peut-être via une autre formule d'intégration ?
        self.state += self.poly*block
        self.output = self.state[:self.block_size]
        self.state >>= self.block_size


class DivByPoly(StreamOperation):

    def __init__(self, series, poly):
        assert poly.parent() is series.parent
        super().__init__(poly.parent(), arity=1, block_size=poly.degree() + 1)
        series.subscribe(self)
        self.poly = poly  # just for printing
        # precomputed inverse: poly*inv = 1 + x^n*rem
        self.inv = poly.inverse_series_trunc(self.block_size)
        self.rem = (self.inv*poly) >> self.block_size
        # state
        self.quo = self.parent.zero()
        self.num = self.parent.zero()

    def __repr__(self):
        return f"{type(self).__name__}(deg={self.poly.degree()})"

    def process_block(self, block):  # likely improvable
        # XXX coûteux (normal...)
        self.num += block
        # f/poly = inv*poly - x^n*(rem*f)/poly
        # (f0 + x^n*f1)/poly
        # = low(inv*f0) + x^n*(high(u*f0) + (-rem*f0 + f1)/poly)
        f0 = self.num[:self.block_size]
        self.quo += self.inv*f0
        self.num = (self.num >> self.block_size) - self.rem*f0
        self.output = self.quo[:self.block_size]
        self.quo >>= self.block_size  # high(inv*f0)


class Add(StreamOperation):

    def __init__(self, series):
        assert len(series) > 0
        assert all(ser.parent is series[0].parent for ser in series)
        super().__init__(series[0].parent, arity=len(series),
                         block_size=max(ser.block_size for ser in series))
        for i, ser in enumerate(series):
            ser.subscribe(self, i)

    def process_block(self, *blocks):
        self.output = sum(blocks)


class EvalSum(StreamOperation):

    def __init__(self, series, point):
        super().__init__(series.parent, arity=1, block_size=series.block_size)
        series.subscribe(self)
        self.point = point
        self.point_pow = point.parent().one()
        self.point_pow_block = point**series.block_size
        self.value = point.parent().zero()  # assumes large enough parent

    def push_coefficients(self, coeff, push_len, operand):
        super().push_coefficients(coeff, push_len, operand)

    def process_block(self, block):
        self.value += self.point_pow*block(self.point)
        # XXX redundant computation (DACUnroller + each EvalSum)
        self.point_pow *= self.point_pow_block


# On pourrait essayer d'autres formulations : séparer les multiplications par
# les matrices aux et itnum, utiliser des dénominateurs distincts pour les
# lignes de aux, réécrire la formule par intégration par parties... C'est une
# partie de l'intérêt de tout le bazar ci-dessus, l'autre étant de rendre les
# bouts plus faciles à tester séparément.
#
# 2024-12-17 Voir aussi si une structure des degrés/dénominateurs issue de la
# filtration de Hodge se retrouve quelque part et si on pourrait l'exploiter.
def shifted_int_mat_times_fmat(num, den, pts, Pol, blksz):
    assert len(den) == num.nrows()
    sources = [Source(Pol, blksz) for _ in range(num.ncols())]
    fmatA = [sources]
    for i in range(num.ncols() - 1):
        fmatA.append([Diff(f) for f in fmatA[-1]])
    # XXX manque-t-il une multiplication ou division par une diagonale de
    # fatorielles quelque part ? non, je ne crois pas : dans cette partie
    # symbolique, tout est écrit en termes des dérivées usuelles
    expr = [
        [[EvalSum(
            Int(
                DivByPoly(
                    # on veut peut-être garder les coefficients rationnels dans
                    # les polynômes...
                    Add([MulByPoly(Pol(num[i][k]), fmatA[k][j])
                         for k in range(num.ncols())]),
                    Pol(den[i]))),
            pt)
          for j in range(num.ncols())]
         for i in range(num.nrows())]
        for pt in pts]
    return sources, expr


def uncouple(sys, den, vec=None):
    r"""
    Uncouple a differential system (fraction-free).

    Return:
    * a diff op annihilating ``y = vec·Y`` where ``Y`` is a vector such that
      ``den·Y' = Y``,
    * a matrix ``T(x)`` s.t. ``T(x)·Y(x) = [y, y', y'', ...]``.
    """
    # maybe clear denominators from the coefficients of sys and den
    Mat = sys.parent()
    Pol = sys.base_ring()
    Dop = OreAlgebra(Pol, 'D' + Pol.variable_name())
    if vec is None:
        vec = Mat.column_space().gen(0)
    else:
        vec = Mat.column_space()(vec)
    x = Pol.gen()
    dden = den.derivative(x)
    mat = matrix(vec)
    for k in range(sys.ncols()):
        v = mat.row(-1)
        v1 = den*v.derivative(x) + v*sys - k*dden*v
        mat = mat.stack(v1)
    denpow = Pol.one()
    for i in range(mat.nrows() - 2, -1, -1):
        denpow *= den
        mat.rescale_row(i, denpow)
    # assert mat.row(0) == den**sys.ncols()*vec
    #ker = mat.left_kernel_matrix()
    #ker = mat.minimal_kernel_basis()  # LEFT kernel
    solver = Dop._solver(mat.base_ring())
    ker = matrix(solver(mat.transpose()))
    if ker.nrows() != 1:
        raise RuntimeError("not cyclic")
    dop = Dop.change_ring(Frac(Pol.change_ring(ZZ)))(list(ker[0])).numerator()
    # maybe keep num&den separate in transformation matrix
    return dop, ~denpow*mat[:-1][:]


class PostIntObserver:

    def __init__(self, num, den, pts):
        self.num = num
        self.den = den
        self.pts = pts
        self.sources = None
        self.expr = None

    def reset(self, unr):
        # TODO better choice of prec here? we *are* losing a significant number
        # of digits in this phase too
        Pol = self.num.base_ring().change_ring(ComplexBallField(unr.sums_prec))
        self.sources, self.expr = shifted_int_mat_times_fmat(
            self.num, self.den, self.pts, Pol, unr.blksz)

    def push_block(self, unr, data):
        for sol, source in zip(data, self.sources):
            assert len(sol) == 1  # log_prec
            source.generate(sol[0])

    def finalize(self):  # XXX un peu pourri comme interface
        for source in self.sources:
            source.close_input()


def fundamental_matrices(sys, den, aux, auxden, path, eps, vec=None, ctx=dctx):

    z = den.parent().gen()
    logger.info("uncoupling...")
    t0 = time.time()
    dop, transf = uncouple(sys, den, vec)
    logger.info("done uncoupling, degree=%s, time=%ss",
                dop.degree(), time.time() - t0)
    dop = DifferentialOperator(dop)
    logger.info("computing transformation...")
    t0 = time.time()
    # wasteful
    itnum, itden = clear_denominators(transf.inverse().list())
    itnum = matrix(transf.nrows(), transf.ncols(), itnum)
    # compute these while we are (presumably) working over QQ
    # XXX pas sûr que ça soit vraiment ce qu'on veut
    # (ni finalement que ça n'ait une importance maintenant qu'on fait le shift
    # numériquement)
    aux1 = aux*itnum
    auxden1 = auxden*itden
    logger.info("done, time=%s s", time.time() - t0)
    dop2 = DifferentialOperator(auxden1*dop)
    # print(len(dop._singularities()), len(dop2._singularities()))
    path = Path(path, dop2)
    # path = Path(path, dop)
    logger.info("path = %ss", path)
    eps = RBF(eps)
    # FIXME prec currently needs to be >= sums_prec (as chosen by HSM) or we
    # are wasting precision
    prec0 = prec_from_eps(eps)
    prec = 6*(prec0 + 2*dop.order())//5 + 100
    Val = ComplexBallField(prec)
    diag = diagonal_matrix(ZZ(i).factorial() for i in range(sys.nrows()))

    logger.info("computing singularities, subdividing path...")
    t0 = time.time()
    # Il peut arriver qu'on se retrouve avec un opérateur singulier en 0
    # même si au départ le système ne l'était pas...
    path = path.bypass_singularities()
    # path.check_singularity()
    path = path.subdivide(1)  # TODO support 2-point mode
    path = path.simplify_points_add_detours(ctx)
    path.check_singularity()
    path.check_convergence()
    logger.info("done, time=%ss, path = %s", time.time() - t0, path)

    tmat_path = block_matrix([[1, MatrixSpace(Val, aux.nrows(), aux.ncols()).zero()],
                              [0, 1]])
    steps = list(path.steps())
    steps.reverse()
    while steps:

        step = steps.pop()
        logger.info("step %s", step)
        t0 = time.time()
        evpts = EvaluationPoint_step([step], jet_order=sys.nrows())
        deltas = [evpts.approx(Val, i) for i in range(len(evpts))]
        z0 = Val(step.start.as_sage_value())
        z1 = Val(step.end.as_sage_value())

        # TODO bornes d'erreur sur la partie aux
        # il faut, en gros :
        # - bien comprendre ce qu'on fait exactement avec les derniers
        # coefficients des développements en série des intégrales (toutes les
        # séries en jeu n'étant pas tronquées au même ordre, on se retrouve
        # peut-être par exemple avec des sommes de trucs tronqués à des ordres
        # différents)
        # - multiplier le majorant sur le reste de la sys par un majorant
        # rationnel convenable
        post_integrator = PostIntObserver(
            aux1(z0+z),
            [auxden1(z0+z)]*aux.nrows(),
            deltas)
        ldop = dop.shift(step.start)
        ctx = Context(ctx=ctx)
        ctx._set_interval_fields(256)
        ctx.__coeff_observer=post_integrator
        hsm = HighestSolMapper_dac(ldop, evpts, eps, fail_fast=True,
                                   effort=0,  # ???
                                   ctx=ctx)
        try:
            cols = hsm.run()
        except (BoundPrecisionError, PrecisionError):
            steps.extend(reversed(step.split()))
            continue

        # XXX redundant with adjacent steps
        transf0 = transf(z0)
        transf1 = transf(z1)

        fmats = []
        for m, delta in enumerate(deltas):  # only one iteration for now
            tmat_dop = matrix([sol.value[m] for sol in cols]).transpose()
            vmat_sys = ~transf1*diag*tmat_dop*~diag*transf0
            vmat_aux = matrix([[post_integrator.expr[m][i][j].value
                                for j in range(aux.ncols())]
                               for i in range(aux.nrows())])
            vmat_aux = vmat_aux*~diag*transf0
            fmat = block_matrix([[1, vmat_aux], [0, vmat_sys]])
            fmats.append(fmat)

        # XXX this is certainly improvable...
        if any(c.rad() > 2.**(-prec0) and c.accuracy() < prec0//2
               for c in vmat_aux.list()):
            steps.extend(reversed(step.split()))
            continue

        [fmat] = fmats
        tmat_path = fmat*tmat_path

        logger.info("done with step %s, time=%ss", step, time.time() - t0)

    return tmat_path



# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sage.rings.complex_arb import ComplexBallField
from sage.rings.complex_mpfr import ComplexField

class Context(object):

    def __init__(self,
            method=None,
            singular=False,
            debug=False,
            use_symmetry=True,
            nbits=200,
            long_fibration=True,
            depth=0,
            simultaneous_integration=False
        ):
        r"""
        Lefschetz Family integration context

        Options:

        * ``method`` -- The way the paths are computed, either along a Voronoi diagram of the singularities ("voronoi"), or a Delaunay triangulation of the singularities ("delaunay"). Default is "voronoi"
        * ``compute_periods`` -- Whether the algorithm should compute periods of the variety, or stop at homology. Default is True.
        * ``singular`` -- Whether the input variety is expected to be singular. Default is False

        * (other options still to be documented...)
        """

        if not method in [None, "voronoi", "delaunay_dual"]:
            raise ValueError("method", method)
        self.method = "voronoi" if method==None else method

        if not isinstance(singular, bool):
            raise TypeError("singular", type(singular))
        self.singular = singular
        
        if not isinstance(debug, bool):
            raise TypeError("debug", type(debug))
        self.debug = debug

        if not isinstance(long_fibration, bool):
            raise TypeError("long_fibration", type(debug))
        self.long_fibration = long_fibration

        if not isinstance(simultaneous_integration, bool):
            raise TypeError("simultaneous_integration", type(debug))
        self.simultaneous_integration = simultaneous_integration

        # if not isinstance(nbits, ): # what type is int ?
        #     raise TypeError("nbits", type(nbits))
        self.nbits = nbits

        if not isinstance(use_symmetry, bool):
            raise TypeError("use_symmetry", type(use_symmetry))
        self.use_symmetry = use_symmetry

        # if not isinstance(depth, int):
        #     raise TypeError("depth", type(depth))
        # self.depth = depth

        self.CBF = ComplexBallField(4*nbits)
        self.CF = ComplexField(4*nbits)
        self.depth = depth
        self.cutoff_simultaneous_integration = 2

dctx = Context() # default context



# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from ore_algebra import *

from sage.matrix.special import identity_matrix
from sage.parallel.decorate import parallel
from ore_algebra.analytic.differential_operator import DifferentialOperator

from sage.rings.integer_ring import Z

from .util import Util

import logging
import os
import time

logger = logging.getLogger(__name__)


class Integrator(object):
    def __init__(self, path_structure, operator, nbits):
        self._operator = DifferentialOperator(operator)
        self.operator._singularities()
        self.nbits = nbits
        self.voronoi = path_structure

    @property
    def operator(self):
        return self._operator
    

    @property
    def transition_matrices(self):
        if not hasattr(self, "_transition_matrices"):
            transition_matrices = []
            for path in self.voronoi.pointed_loops:
                transition_matrix = 1
                N = len(path)
                for i in range(N-1):
                    e = path[i:i+2]
                    if e in self.voronoi.edges:
                        index = self.voronoi.edges.index(e)
                        transition_matrix = self.integrated_edges[index] * transition_matrix
                    else:
                        index = self.voronoi.edges.index([e[1], e[0]])
                        transition_matrix = self.integrated_edges[index]**-1 * transition_matrix
                transition_matrices += [transition_matrix]
            self._transition_matrices = transition_matrices
        return self._transition_matrices

    def find_complex_conjugates(self):
        complex_conjugates = [None]*len(self.voronoi.vertices)
        for i in range(len(self.voronoi.vertices)):
            if complex_conjugates[i]==None:
                if self.voronoi.vertices[i].conjugate() in self.voronoi.vertices:
                    complex_conjugates[i] = self.voronoi.vertices.index(self.voronoi.vertices[i].conjugate())
                    complex_conjugates[complex_conjugates[i]] = i
        return complex_conjugates

    @property
    def integrated_edges(self):
        if not hasattr(self, "_integrated_edges"):
            complex_conjugates = self.find_complex_conjugates()
            index_of_edges_to_integrate = []
            edges=[]
            for i, e in enumerate(self.voronoi.edges):
                if [e[1], e[0]] not in edges and [complex_conjugates[e[0]], complex_conjugates[e[1]]] not in edges and [complex_conjugates[e[1]], complex_conjugates[e[0]]] not in edges:
                    index_of_edges_to_integrate+=[i]
                    edges+=[e]

            edges = [[self.voronoi.vertices[e[0]], self.voronoi.vertices[e[1]]] for e in edges]
            N = len(edges)
            integration_result = Integrator._integrate_edge([([i,N],self.operator,[e[0], e[1]], self.nbits) for i, e in list(enumerate(edges))])
            integrated_edges_temp= [None]*N

            for [inp, _], ntm in integration_result:
                if ntm == 'NO DATA': # why is this not a result of @parallel?
                    logger.warning("failed to integrate operator")
                integrated_edges_temp[inp[0][0]] = ntm # there should be a cleaner way to do this

            integrated_edges = [None]*len(self.voronoi.edges)
            for index, i in enumerate(index_of_edges_to_integrate):
                integrated_edges[i] = integrated_edges_temp[index]
                e = self.voronoi.edges[i]
                if [complex_conjugates[e[0]], complex_conjugates[e[1]]] == e:
                    continue
                if [complex_conjugates[e[0]], complex_conjugates[e[1]]] in self.voronoi.edges:
                    j = self.voronoi.edges.index([complex_conjugates[e[0]], complex_conjugates[e[1]]])
                    integrated_edges[j] = integrated_edges_temp[index].conjugate()
                if [complex_conjugates[e[1]], complex_conjugates[e[0]]] in self.voronoi.edges:
                    j = self.voronoi.edges.index([complex_conjugates[e[1]], complex_conjugates[e[0]]])
                    integrated_edges[j] = integrated_edges_temp[index].inverse().conjugate()

            self._integrated_edges = integrated_edges
        return self._integrated_edges
    
    @classmethod
    @parallel
    def _integrate_edge(cls, i, L, l, nbits=300, maxtries=5, verbose=False):
        """ Returns the numerical transition matrix of L along l, adapted to computations of Voronoi. Accepts l=[]
        """
        logger.info("[%d] Starting integration along edge [%d/%d]"% (os.getpid(), i[0]+1,i[1]))
        tries = 1
        bounds_prec=256
        begin = time.time()
        eps = Z(2)**(-Z(nbits))
        while True:
            # try:
            ntm = L.numerical_transition_matrix(l, eps=eps, assume_analytic=True, bounds_prec=bounds_prec) if l!= [] else identity_matrix(L.order()) 
            ntmi = ntm**-1 # checking the matrix is precise enough to be inverted
            # except Exception as e: # TODO: manage different types of exceptions
            #     tries+=1
            #     if tries<maxtries:
            #         bounds_prec *=2
            #         nbits*=2
            #         logger.info("[%d] Precision error when integrating edge [%d/%d]. Trying again with double bounds_prec (%d) and nbits (%d)."% (os.getpid(), i[0]+1, i[1], bounds_prec, nbits))
            #         continue
            #     else:
            #         logger.info("[%d] Too many ValueErrors when integrating edge [%d/%d]. Stopping computation here"% (os.getpid(), i[0]+1, i[1]))
            #         raise e
            break

        end = time.time()
        duration = end-begin
        duration_str = time.strftime("%H:%M:%S",time.gmtime(duration))
        logger.info("[%d] Finished integration along edge [%d/%d] in %s"% (os.getpid(), i[0]+1,i[1], duration_str))

        return ntm


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from .numperiods.family import Family
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.qqbar import QQbar
from sage.functions.other import factorial
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.misc.misc_c import prod

from .voronoi import FundamentalGroupVoronoi
from .integrator import Integrator
from .util import Util
from .context import Context
from .hypersurface import Hypersurface
from .monodromyRepresentation import MonodromyRepresentation

import logging
import time

logger = logging.getLogger(__name__)


class Fibration(object):
    def __init__(self, P, basepoint=None, fibration=None, fibre=None, cyclic_form = None, **kwds):
        """P, a homogeneous polynomial defining a family of hypersurfaces.
        """
        
        self.ctx = Context(**kwds)

        self._P = P
        self._fibration = fibration
        
        _, denom = Family(self.P, path=[-1, 0]).gaussmanin()
        self._critical_values = denom.roots(QQbar, multiplicities=False)

        if cyclic_form!= None: 
            L = self.family.picard_fuchs_equation(cyclic_form)
            assert L.order() == len(self.fibre.cohomology), "cyclic_form is not cyclic"
            self._cyclic_form = cyclic_form
            self._cyclic_picard_fuchs_equation = L
        
        self._family = Family(self.P, basepoint=basepoint)

        if fibre!=None:
            assert basepoint!=None, "Cannot specify fibre without specifying basepoint"
            assert P(basepoint) == fibre.P, "Fibre and P(basepoint) do not match"
            self._fibre = fibre

        if basepoint != None: # it is useful to be able to specify the basepoint to avoid being stuck in arithmetic computations if critical values have very large modulus
            assert basepoint not in self.critical_values, "basepoint is not regular"
            self._basepoint = basepoint
        
        if not self.ctx.debug:
            fg = self.fundamental_group # this allows reordering the critical points straight away and prevents shenanigans. There should be a better way to do this
        
    
    @property
    def monodromy_representation(self):
        if not hasattr(self,'_monodromy_representation'):
            self._monodromy_representation = MonodromyRepresentation(self.monodromy_matrices, self.fibre.intersection_product)
        return self._monodromy_representation
    
    @property
    def intersection_product(self):
        return self.monodromy_representation.intersection_product

    @property
    def P(self):
        return self._P

    @property
    def picard_fuchs_equations(self):
        if not hasattr(self,'_picard_fuchs_equations'):
            self._picard_fuchs_equations = [self.family.picard_fuchs_equation(vector([w,0])) for w in self.holomorphic_forms]
        return self._picard_fuchs_equations
    
    @property
    def family(self):
        return self._family
    
    @property
    def critical_values(self):
        if not hasattr(self,'_critical_values'):
            _, denom = self.family.gaussmanin
            self._critical_values = denom.roots(QQbar, multiplicities=False)
        return self._critical_values

    def vector_to_form(self, v):
        dmax = max([w.degree() for w in self.fibre.cohomology])
        dim = self.fibre.dim
        deg = self.fibre.degree
        res = 0
        smax = (dmax + dim + 1) // deg
        for c, w in zip(v, self.fibre.cohomology):
            s = (w.degree() + dim + 1) // deg
            res += self.P.parent()(factorial(s) * self.P**(smax-s) * w) * self.P.parent(c)
        return res


    @property
    def monodromy_matrices(self):
        if not hasattr(self, '_monodromy_matrices'):
            r = len(self.fibre.cohomology) 
            
            cyclic_form = self.cyclic_form
            w = self.vector_to_form(cyclic_form)

            integration_correction = diagonal_matrix([1]+[1/ZZ(factorial(k)) for k in range(r)])
            derivatives_at_basepoint = self.derivatives_values_at_basepoint(w)
            cohomology_fibre_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fibre.cohomology], self.basepoint)

            initial_conditions = integration_correction * derivatives_at_basepoint * cohomology_fibre_to_family.inverse()
            initial_conditions = initial_conditions.submatrix(1,0)

            cohomology_monodromies = [initial_conditions.inverse() * M * initial_conditions for M in self.cyclic_transition_matrices]
            if self.fibre.dim%2 == 0:
                cohomology_monodromies = [block_diagonal_matrix([M, identity_matrix(1)]) for M in cohomology_monodromies]

            Ms = [(self.fibre.period_matrix.inverse() * M * self.fibre.period_matrix) for M in cohomology_monodromies]

            try:
                Ms = [M.change_ring(ZZ) for M in Ms]
            except Exception as e:
                if not self.ctx.debug:
                    raise e
            
            Mtot = prod(list(reversed(Ms)))
            if Mtot != 1:
                self._critical_values = self.critical_values + ["infinity"]
                transition_matrix_infinity = prod(list(reversed(self.cyclic_transition_matrices))).inverse()
                self._cyclic_transition_matrices += [transition_matrix_infinity]
                Ms += [(Mtot.inverse()).change_ring(ZZ)]
                
                self._paths += [-sum(self.paths)]
            
            self._monodromy_matrices = Ms
        return self._monodromy_matrices
    
    @property
    def thimbles_confluence(self):
        if not hasattr(self, '_thimbles_confluence'):
            blocks =[]
            for i, pcs in enumerate(self.permuting_cycles):
                decompositions = []
                for p in pcs:
                    decomposition = []
                    for M, v in zip(self.monodromy_matrices_morsification[i], self.vanishing_cycles_morsification[i]):
                        decomposition += [(M-1)*p/v]
                        p = M*p
                    decompositions+=[decomposition]
                blocks+= [matrix(decompositions)]
            self._thimbles_confluence = block_diagonal_matrix(blocks).change_ring(ZZ)
        return self._thimbles_confluence

    @property
    def fibre(self):
        if not hasattr(self,'_fibre'):
            self._fibre = Hypersurface(self.P(self.basepoint), nbits=self.ctx.nbits, fibration=self._fibration)
        return self._fibre
    
    @property
    def cyclic_form(self):
        if not hasattr(self, '_cyclic_form'):
            r = len(self.fibre.cohomology)
            t = self.family.upolring.gen(0)
            for v in identity_matrix(r).rows()+[vector([t**i for i in range(r)])]:
                L = self.family.picard_fuchs_equation(v)
                if L.order() == r:
                    break
            assert L.order() == r, "could not find cyclic form"
            self._cyclic_form = v
            self._cyclic_picard_fuchs_equation = L
        return self._cyclic_form
    
    @property
    def cyclic_picard_fuchs_equation(self):
        if not hasattr(self, '_cyclic_picard_fuchs_equation'):
            self.cyclic_form
        return self._cyclic_picard_fuchs_equation
    
    @property
    def cyclic_transition_matrices(self):
        if not hasattr(self, '_cyclic_transition_matrices'):
            L = self.cyclic_picard_fuchs_equation
            self._cyclic_transition_matrices = self.integrate(L)
        return self._cyclic_transition_matrices
    
    def integrate(self, L):
        logger.info("Computing numerical transition matrices of operator of order %d and degree %d (%d edges total)."% (L.order(), L.degree(), len(self.fundamental_group.edges)))
        begin = time.time()

        integrator = Integrator(self.fundamental_group, L, self.ctx.nbits)
        transition_matrices = integrator.transition_matrices
        
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("Integration finished -- total time: %s."% (duration_str))

        return transition_matrices


    def derivatives_values_at_basepoint(self, w):
        s = len(self.fibre.cohomology)
        derivatives = [self.P.parent()(0), w]
        for k in range(s-1):
            derivatives += [self._derivative(derivatives[-1], self.P)] 
        return self.family._coordinates(derivatives, self.basepoint)

    @classmethod
    def _derivative(self, A, P): 
        """computes the numerator of the derivative of A/P^k"""
        field = P.parent().fraction_field()
        return field(A).derivative() - A*P.derivative()         

    @property
    def fundamental_group(self):
        if not hasattr(self,'_fundamental_group'):
            begin = time.time()

            fundamental_group = FundamentalGroupVoronoi(self.critical_values, self.basepoint) # access future delaunay implem here
            fundamental_group.sort_loops()

            end = time.time()
            duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
            logger.info("Fundamental group computed in %s."% (duration_str))

            self._critical_values = fundamental_group.points[1:]
            self._fundamental_group = fundamental_group
        return self._fundamental_group

    @property
    def paths(self):
        if not hasattr(self,'_paths'):
            paths = []
            for path in self.fundamental_group.pointed_loops:
                paths += [[self.fundamental_group.vertices[v] for v in path]]
            self._paths= paths
        return self._paths

    @property
    def basepoint(self):
        if  not hasattr(self, '_basepoint'):
            shift = 1
            reals = [self.ctx.CF(c).real() for c in self.critical_values]
            xmin, xmax = min(reals), max(reals)
            self._basepoint = Util.simple_rational(xmin - (xmax-xmin)*shift, (xmax-xmin)/10)
        return self._basepoint



# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.qqbar import QQbar
from sage.matrix.special import zero_matrix

from .numperiods import interpolation

from sage.rings.imaginary_unit import I
from sage.groups.free_group import FreeGroup

from .voronoi import FundamentalGroupVoronoi
from .util import Util
from .translator import Translator
from .rootsBraid import RootsBraid


import logging
import time

logger = logging.getLogger(__name__)


class ExceptionalDivisorComputer(object):
    def __init__(self, variety):
        """
        """
        self.variety = variety
        self.Qt = PolynomialRing(QQ, 't')
        self.Qu = PolynomialRing(QQ, ['u', 't'])

    def _compute_coefs(self, b):
        R = self.variety.family.pol(b).parent()
        _vars = [v for v in R.gens()]
        forms=[v.dot_product(vector(_vars)) for v in self.variety.fibre.fibration[:2]]
        f=forms[0]/forms[1]
        S = PolynomialRing(QQ, _vars+['k','t'])
        k,t= S.gens()[-2:]
        eqs = [
            self.variety.family.pol(b), 
            forms[1]-1, 
            t*forms[1]-forms[0]
        ] + [(f.derivative(var).numerator()*k-self.variety.family.pol(b).derivative(var)*f.derivative(var).denominator()) for var in _vars]

        ideal = S.ideal(eqs).elimination_ideal(S.gens()[:-1])
        Qt = PolynomialRing(QQ, 't')

        return Qt(ideal.groebner_basis()[0]).coefficients(sparse=False)
    
    @property
    def critical_values_polynomial(self):
        if not hasattr(self, "_critical_values_polynomial"):
            u, t = self.Qu.gens()
            fr = interpolation.FunctionReconstruction(self.Qt, self._compute_coefs)
            coefs, denom = fr.recons(denomapart=True)
            self._critical_values_polynomial = sum([c(u)*t**i for i,c in zip(range(len(coefs)),coefs)])
        return self._critical_values_polynomial
    
    @property
    def fundamental_group_critical(self):
        """This is the fundamental group of C punctured at the points where the critical polynomial has multiple roots."""
        if not hasattr(self, "_fundamental_group_critical"):
            u, t = self.Qu.gens()
            double_roots = self.Qt((self.critical_values_polynomial*(t-self.variety.fibre.basepoint)).discriminant(t)(u=t)).roots(QQbar, multiplicities=False)
            fg= FundamentalGroupVoronoi(double_roots,self.variety.basepoint)
            fg.sort_loops()
            self._fundamental_group_critical = fg
        return self._fundamental_group_critical
    
    @property
    def adapted_paths(self):
        if not hasattr(self, "_adapted_paths"):
            translator = Translator(self.variety.fundamental_group,self.fundamental_group_critical)
            adapted_paths = [translator.specialize_path(path) for path in self.variety.fundamental_group.pointed_loops]

            # the following lines are here to remove loops around infinity. They are not really mandatory.
            looptot = []
            for path in adapted_paths:
                looptot += path
            looptot = Util.simplify_path(looptot)
            while looptot[0] == looptot[-1]:
                v = looptot[0]
                looptot = looptot[1:-1]
            looptot = [v]+looptot +[v]

            n = len(looptot)
            for j, path in enumerate(adapted_paths):
                for i in range(len(path)-n, 0, -1):
                    if path[i:i+n] in [looptot, list(reversed(looptot))]:
                        path = path[:i] + path[i+n-1:]
                adapted_paths[j] = path

            self._adapted_paths = adapted_paths
        return self._adapted_paths
    
    @property
    def edges(self):
        if not hasattr(self, "_edges"):
            _edges = []
            for path in self.adapted_paths:
                for i in range(len(path)-1):
                    e = path[i:i+2]
                    if e not in _edges and list(reversed(e)) not in _edges: 
                        _edges += [e]
            edges_spec = [[self.fundamental_group_critical.vertices[c] for c in e] for e in _edges]
            self._edges = edges_spec
        return self._edges

    @property
    def roots_braid(self):
        if not hasattr(self, "_roots_braid"):
            self._roots_braid = RootsBraid(self.critical_values_polynomial, self.edges, additional_points=[QQ(self.variety.fibre.basepoint)])
        return self._roots_braid

    @property
    def marking_init(self):
        """This is the initial marking at the basepoint"""
        if not hasattr(self, "_marking_init"):
            i = Util.select_closest_index(self.roots_braid.vertices, self.fundamental_group_critical.points[0])
            self._marking_init = self.roots_braid.system(i) + self.roots_braid.additional_points
        return self._marking_init
    
    @property
    def fundamental_group_fibre(self):
        if not hasattr(self, "_fundamental_group_fibre"):
            xmax=Util.simple_rational(max([s.real() for s in self.marking_init]), 0.000001)
            xmin=Util.simple_rational(min([s.real() for s in self.marking_init]), 0.000001)
            ymax=Util.simple_rational(max([s.imag() for s in self.marking_init]), 0.000001)
            fake_basepoint = 2*xmin-xmax+ymax*I/5
            fundamental_group_fibre = FundamentalGroupVoronoi(self.variety.fibre.critical_values + [self.variety.fibre.basepoint], fake_basepoint)
            fundamental_group_fibre.sort_loops()
            self._fundamental_group_fibre = fundamental_group_fibre
        return self._fundamental_group_fibre
    
    @property
    def words_fibre(self):
        if not hasattr(self, "_words_fibre"):
            fibre_translator = Translator(self.variety.fibre.fundamental_group, self.fundamental_group_fibre)
            spec_f = [fibre_translator.specialize_path(path) for path in self.variety.fibre.fundamental_group.pointed_loops]
            fakebp = self.fundamental_group_fibre.points[0]
            
            s_to_FG = [Util.select_closest_index(self.fundamental_group_fibre.points, s) for s in self.marking_init+[fakebp]]
            edges = [list(e[:2]) for e in self.roots_braid.minimal_cover_tree(self.marking_init).edges()]
            for i, e in enumerate(edges): # orient the edges away from basepoint
                if self.roots_braid.minimal_cover_tree(self.marking_init).distance(e[1], self.roots_braid.npoints) < self.roots_braid.minimal_cover_tree(self.marking_init).distance(e[0],self.roots_braid.npoints):
                    edges[i] = list(reversed(e))
            mtc_init = [[s_to_FG[i] for i in e] for e in edges]

            edges = [[c,p] for c, p in self.fundamental_group_fibre.polygons if c==self.variety.fibre.basepoint][0][1]
            cycle = [i for i in edges[0]]
            while cycle[0] != cycle[-1]:
                for e in edges:
                    if cycle[-1] in e and cycle[-2] not in e:
                        if e[0] == cycle[-1]:
                            cycle+=[e[1]]
                        else:
                            cycle+=[e[0]]
                        break
            cycle = list(reversed(cycle)) if Util.is_clockwise([self.fundamental_group_fibre.vertices[i] for i in cycle[:-1]]) else cycle
            cycle = cycle[:-1]

            for path in spec_f:
                for v in path:
                    if v in cycle:
                        i = cycle.index(v)
                        j = path.index(v)
                        cycle = cycle[i:] + cycle[:i]
                        res = path[:j] + cycle + list(reversed(path[:j+1]))
                        break
                else:
                    continue
                break
            spec_f = [res] + spec_f
            self._spec_f  = spec_f # it's nice to have this for plotting purposes

            self._words_fibre = [Translator.word(path, self.fundamental_group_fibre.duality, mtc_init, self.roots_braid.freeGroup) for path in spec_f]

        return self._words_fibre 
    
    @property
    def thimble_monodromy(self):
        if not hasattr(self, "_thimble_monodromy"):
            thimblegroup = FreeGroup(len(self.words_fibre), 't')
            ts = thimblegroup.gens()
            ttox = thimblegroup.hom(self.words_fibre)
            xtot = Util.invert_morphism(ttox)
            
            isos = []
            thimble_monodromy = []
            adapted_paths_z = [[self.fundamental_group_critical.vertices[v] for v in path] for path in self.adapted_paths]
            
            begin=time.time()
            logger.info("Computing the braid action.")
            self.roots_braid.compute_all_isomorphisms()
            logger.info("There are %d edges in total."% len(self.roots_braid.edges))
            for index_thimble, thimble_path in enumerate(adapted_paths_z):
                logger.info("Computing monodromy of path between blowups along loop %d/%d: %d edges "% (index_thimble+1,len(adapted_paths_z), len(thimble_path)))
                iso = self.roots_braid.isomorphism_along_path(thimble_path)
                isos += [iso]
                monodromy = zero_matrix(len(ts)-1)
                
                conjtobp = Util.middle(xtot(iso(ttox(ts[0]))))
                
                for tinit, chain in zip(ts[1:], self.variety.fibre.thimbles):
                    v = chain[0]
                    ttilde = xtot(iso(ttox(tinit)))
                    ttilde = conjtobp**-1*ttilde*conjtobp
                    
                    j = ts.index(tinit)-1
                    for t, p in list(reversed(ttilde.syllables())):
                        assert p in [-1,1]
                        i = ts.index(t)
                        if i==0:
                            continue
                        i-=1
                        monodromy[i, j]+=(self.variety.fibre.monodromy_matrices[i]**p-1)*v / self.variety.fibre.vanishing_cycles[i]
                        v = self.variety.fibre.monodromy_matrices[i]**p*v
                thimble_monodromy += [monodromy]
                assert v-chain[0] == self.variety.fibre.vanishing_cycles[j], "boundaries not matching"
            self._thimble_monodromy = thimble_monodromy
            self._isos = isos
            end = time.time()
            duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
            logger.info("Braid action computed in %s.", duration_str)
        return self._thimble_monodromy

            


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from sage.arith.misc import gcd
from sage.rings.integer_ring import ZZ

from .monodromyRepresentation import MonodromyRepresentation

import logging
import time

logger = logging.getLogger(__name__)


class MonodromyRepresentationSurface(MonodromyRepresentation):

    def desingularise_matrix(self, M):
        if (M-1).rank() != 1:
            raise Exception("Unknown singular fibre type")
        v = (M-1).image().gen(0)
        n = gcd(v)
        decomposition = [(M-1)/n + 1] * n
        decomposition = [M.change_ring(ZZ) for M in decomposition]
        return decomposition

    @property
    def self_intersection_section(self):
        if not hasattr(self, '_self_intersection_section'):
            self._self_intersection_section = -1
        return self._self_intersection_section

    @property
    def add(self):
        if not hasattr(self, '_add'):
            self._add = 2
        return self._add


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from ore_algebra import *

from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.qqbar import QQbar
from sage.graphs.graph import Graph
from sage.rings.imaginary_unit import I

from sage.rings.complex_mpfr import ComplexField
from sage.groups.free_group import FreeGroup
from sage.misc.flatten import flatten
from sage.schemes.curves.zariski_vankampen import followstrand
from sage.functions.other import arg

from sage.parallel.decorate import parallel


from .util import Util

import logging
import time
import os
from copy import copy

logger = logging.getLogger(__name__)


class RootsBraid(object):
    def __init__(self, P, edges, basepoint=None, additional_points=[]):
        """P, a polynomial in two variables u and t.

        This class computes the braid group of roots (in t) of P(u) as u moves along a path
        """
        
        R = PolynomialRing(QQ, ['u','t'])
        u,t = R.gens()
        self.P = P(u,t)
        self.additional_points=additional_points
        self.npoints = self.P.degree(self.P.parent().gens()[1]) + len(self.additional_points)
        self._maximalstep = 1/1000

        self.freeGroup = FreeGroup(self.npoints)
        self.xs = list(self.freeGroup.gens())
        # self.additional_points=additional_points

        self.hasbasepoint = (basepoint != None)
        self.basepoint = basepoint

        self.edges = []
        self.vertices = []
        for e in edges:
            if e[0] not in self.vertices:
                self.vertices += [e[0]]
            if e[1] not in self.vertices:
                self.vertices += [e[1]]
            e2 = [self.vertices.index(v) for v in e]
            if e2 not in self.edges and list(reversed(e2)) not in self.edges:
                self.edges+=[e2]



    @property
    def singularities(self):
        if not hasattr(self,'_singularities'):
            t = self.P.parent()('t')
            discrP = self.P.discriminant(t)
            Qu=PolynomialRing(QQ[I], 'u')
            self._singularities = Qu(discrP).roots(QQbar, multiplicities=False)
        return self._singularities
    
    def system(self, i):
        """The configuration of roots at self.vertices[i]"""
        if not hasattr(self,'_systems'):
            self._systemsQ = [False]*len(self.vertices)
            self._systems = [None]*len(self.vertices)
        if not self._systemsQ[i]:
            CC=ComplexField(500)
            Qt=PolynomialRing(QQ[I], 't')
            p = self.vertices[i]
            u,t = self.P.parent()('u'),self.P.parent()('t')
            roots = Qt(self.P(u=p)).roots(QQbar, multiplicities=False)
            roots = [CC(r)for r in roots]
            roots.sort(key=lambda z: (z.real(), z.imag()))
            self._systems[i] = roots
            self._systemsQ[i] = True
        return self._systems[i]

    def compute_all_braids(self):
        if not hasattr(self,'_braid'):
            self._braid = [(None, None)]*len(self.edges)
            self._braidQ = [False]*len(self.edges)
        begin = time.time()
        logger.info("Computing all braids (%d in total).", (len(self.edges)))
        result = self._compute_braid([(e,i) for i, e in enumerate(self.edges) if not self._braidQ[i]])
        for arg, res in result:
            braid, braidinverse = res
            i, _ =  self.edge(arg[0][0])
            self._braid[i] = braid, braidinverse
            self._braidQ[i] = True
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("Braids computed in %s."% (duration_str))

    def braid(self, e):
        if not hasattr(self,'_braid'):
            self._braid = [(None, None)]*len(self.edges)
            self._braidQ = [False]*len(self.edges)

        i, inverse = self.edge(e)
        if not self._braidQ[i]:
            logger.info("[%d] Computing braid along edge %d."% (os.getpid(), i))

            res, resinverse = self._compute_braid(e)
            
            self._braid[i] = (resinverse, res) if inverse else (res, resinverse)
            self._braidQ[i] = True

        return self._braid[i][1 if inverse else 0]

    @parallel
    def _compute_braid(self, e, i):
        from sage.schemes.curves.zariski_vankampen import followstrand
        logger.info("[%d] Computing braid along edge %d"% (os.getpid(), i))
        begin = time.time()
        roots = self.system(e[0])
        res=[]
        j=0
        steps = [begin]
        for r in roots:   
            j+=1
            # logger.info("[%d] Computing thread %d of edge %d."% (os.getpid(), j, i))
            line = followstrand(self.P, [z.minpoly()(self.P.parent().gens()[1]) for z in self.additional_points], self.vertices[e[0]], self.vertices[e[1]],r, 50)
            res+=  [[[c[0], c[1]+I*c[2]] for c in line]]
            steps += [time.time()]
        
        end = time.time()
        duration = end-begin
        duration_str = time.strftime("%H:%M:%S",time.gmtime(duration))
        duration_str_steps = ' '.join([time.strftime("%H:%M:%S",time.gmtime(steps[i]-steps[i-1])) for i in range(1, len(steps))])
        logger.info("[%d] Finished computation of braid along edge %d. [total time:%s], [per root: %s]."% (os.getpid(), i, duration_str, duration_str_steps))

        resinverse = [list(reversed([[1-t, x] for t, x in thread])) for thread in res]
        rootsinverse = self.system(e[1])
        endthreads = [thread[0][1] for thread in resinverse]
        order = [Util.select_closest_index(endthreads, r) for r in rootsinverse]
        resinverse = [resinverse[i] for i in order]
        return res,resinverse

    def interpolate(self, thread, t):
        if t==1:
            return thread[-1][1]
        for i in range(len(thread)-1,-1,-1):
            if t>=thread[i][0]:
                break
        t0,x0, t1,x1 = thread[i][0], thread[i][1], thread[i+1][0], thread[i+1][1]
        return ((t1-t)*x0+ (t-t0)*x1)/(t1-t0)


    def minimal_cover_tree(self, section):
        CC=ComplexField(500)
        mtc=Graph(self.npoints+1) 
        edges = flatten([[(i,j) for i in range(j)] for j in range(self.npoints)], max_level=1) 
        edges.sort(key=(lambda e: Util.simple_rational(abs(CC(section[e[0]]-section[e[1]])), 10e-10))) # we sort edges by length
        for e in edges:
            if len(mtc.shortest_path(e[0], e[1]))==0:
                mtc.add_edge(e)
        # then we add the path to the basepoint
        vertices = [i for i in range(self.npoints)]
        if self.hasbasepoint:
            vertices.sort(key=(lambda v: Util.simple_rational(abs(CC(section[v]-self.basepoint)), 10e-10)))
        else:
            vertices.sort(key=(lambda v: (section[v].real(), -section[v].imag()))) # the fixed basepoint is also specific to example
        mtc.add_edge([self.npoints,vertices[0]])
        return mtc

    def edge(self, e):
        """Given an edge e, returns (i, False) if self.edges[i]==e and (i, True) if self.edges[i]==list(reversed(e))"""
        if [e[0], e[1]] in self.edges:
            return self.edges.index([e[0], e[1]]), False
        if [e[1], e[0]] in self.edges:
            return self.edges.index([e[1], e[0]]), True
        raise Exception("edge is not in edge list")

    def braid_section(self, braid, t):
        """Given a braid (a list of threads) and 0<t<1, returns the section of the braid at t."""
        section = []
        for thread in braid:
            section += [self.interpolate(thread, t)]
        return section+self.additional_points

    def raffine_braid(self, braid):
        ts=[]
        for thread in braid:
            for p in thread:
                ts+=[p[0]]
        ts = list(set(ts))
        ts.sort()
        for t0,t1 in zip(ts[:-1], ts[1:]):
            while t1-t0>1.1*self._maximalstep:
                t0+=self._maximalstep
                ts.append(t0)
        ts.sort()
        return [self.braid_section(braid, t) for t in ts]


    def isomorphisms(self, e):
        if not hasattr(self,'_isomorphisms'):
            self._isomorphisms = [[None, None] for i in range(len(self.edges))]
            self._isomorphismsQ = [False]*len(self.edges)
        i, inverse = self.edge(e)
        if not self._isomorphismsQ[i]:
            logger.info("[%d] Computing isomorphism of edge %d."% (os.getpid(), i))
            CC=ComplexField(500)

            iso = self._compute_isomorphism(e)

            self._isomorphisms[i][1 if inverse else 0] = iso
            self._isomorphismsQ[i] = True

        if self._isomorphisms[i][1 if inverse else 0] == None:
            logger.info("[%d] Inverting isomorphism of edge %d."% (os.getpid(), i))
            self._isomorphisms[i][1 if inverse else 0] = Util.invert_morphism(self._isomorphisms[i][0 if inverse else 1])

        return self._isomorphisms[i][1 if inverse else 0]

    def compute_all_isomorphisms(self):
        if not hasattr(self,'_isomorphisms'):
            self._isomorphisms = [[None, None] for i in range(len(self.edges))]
            self._isomorphismsQ = [False]*len(self.edges)
        self.compute_all_braids()
        begin = time.time()
        logger.info("Computing all isomorphisms.")
        result = self._compute_isomorphism([e for i, e in enumerate(self.edges) if not self._isomorphismsQ[i]])
        for arg, res in result:
            iso = res
            i, _ =  self.edge(arg[0][0])
            self._isomorphisms[i][0] = iso
            self._isomorphismsQ[i] = True
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("Isomorphisms computed in %s."% (duration_str))

    @parallel
    def _compute_isomorphism(self, e):
        i, inverse = self.edge(e)
        braid = self.braid(e)
        sections = self.raffine_braid(braid)

        mtcs = [self.minimal_cover_tree(section) for section in sections]

        mtcinit = self.minimal_cover_tree(self.system(e[0]) + self.additional_points)

        iso = self.freeGroup.hom(self.xs)
        if mtcs[0] != mtcinit:
            logger.info("[%d] Encountered distinct minimal covering trees between beginning of braid %d and standard configuration."% (os.getpid(), i))
            iso = self.braid_action(mtcinit, mtcs[0], sections[0])

        for k in range(len(mtcs)-1):
            mtc1 = mtcs[k]
            mtcfin = mtcs[k+1]
            if mtc1!=mtcfin:
                logger.info("[%d] Encountered distinct minimal covering trees between sections %d and %d (out of %d)."% (os.getpid(), k, k+1, len(mtcs)))
                iso = self.braid_action(mtc1, mtcfin, sections[k])*iso
                iso = self.freeGroup.hom([iso(x) for x in self.xs])

        section1 = sections[-1]
        section2 = self.system(e[1]) + self.additional_points

        perm = [Util.select_closest_index(section2,c) for c in section1]+[self.npoints] # this is fine because they are equal (although their presentation might differ)
            
        mtc1 = mtcs[-1]
        mtcfin = self.minimal_cover_tree(section2)

        mtcn = Graph(self.npoints+1)
        for e in mtc1.edges():
            mtcn.add_edge((perm[e[0]], perm[e[1]]))

        oe1, oe2 = self.ordered_edges(mtc1), self.ordered_edges(mtcn)
        perm_edge = [oe2.index(self.normalize_edge((perm[e[0]],perm[e[1]]))) for e in oe1]
        transition_iso = self.freeGroup.hom([self.xs[i] for i in perm_edge])
        if mtcn!=mtcfin:
            logger.info("[%d] Encountered distinct minimal covering trees between end of braid %d and standard configuration."% (os.getpid(), i))
            transition_iso = self.braid_action(mtcn, mtcfin, section2)*transition_iso

        iso = transition_iso*iso
        iso = self.freeGroup.hom([iso(x) for x in self.xs])
        return iso
    
    def isomorphism_along_path(self,path):
        """Given a path `path`, computes the braid (as an isomorphism on the fundamental group of the punctured plane) along `path`"""
        path = [Util.select_closest_index(self.vertices, p)  for p in path]
        edges = [path[i:i+2] for i in range(len(path)-1)] # TODO try product([self.isomorphism(e) for e in list(reversed(path_edges))])
        
        iso = self.isomorphisms(edges[0])
        for e in edges[1:]:
            iso = self.isomorphisms(e)*iso
            iso = self.freeGroup.hom([iso(x) for x in self.xs])
        return iso

    def edge_difference(self, g1, g2):
        """Given two graphs g1, g2, yields the lists `removed_edges, added_edges` such that `removed_edges` is the edges in `g1` and not in `g2`, and `added_edges` the opposite."""
        removed_edges = []
        added_edges = []
        for e in g1.edges():
            if not g2.has_edge(e):
                removed_edges+=[e]
        for e in g2.edges():
            if not g1.has_edge(e):
                added_edges+=[e]
        return removed_edges, added_edges
    
    def braid_action(self,g1,g2, section):
        """ computes the isomorphism elements characterizing the change from graph g1 to graph g2
        """
        iso = self.freeGroup.hom(self.xs)
        removed_edges, added_edges = self.edge_difference(g1, g2)

        # logger.info("Starting from graph with edges (%d, %d), (%d, %d), (%d, %d)."% tuple(flatten([[e[0], e[1]] for e in g1.edges()])))
        for e in removed_edges:
            # logger.info("Removing edge (%d,%d)."%(e[0], e[1]))
            # we add the edge to see what cycle appears
            gx = copy(g1)
            gx.delete_edge(e)

            if not e[0] in gx.connected_component_containing_vertex(self.npoints):
                e = (e[1], e[0])

            for ea in added_edges:
                ga = copy(gx)
                ga.add_edge(ea)
                if ga.connected_components_number()==1:
                    break
            ga = copy(g1)
            ga.add_edge(ea)
            assert ga.connected_components_number()==1
            # logger.info("Adding edge (%d,%d)."%(ea[0], ea[1]))
            ea = self.normalize_edge(ea)

            cycle = ga.cycle_basis()[0]
            # then we normalize the cycle (we want to start at the vertex from the edge that was deleted that's in the same component as the basepoint)

            ci = cycle.index(e[0])
            if cycle[ci+1 if ci+1<len(cycle) else 0] == e[1]:

                cycle.reverse()
            ci = cycle.index(e[0])
            cycle = cycle[ci:] + cycle[:ci]

            if self.hasbasepoint:
                clockwise = Util.is_clockwise([section[i] if i!=self.npoints else self.basepoint for i in cycle])
            else:
                xmax=Util.simple_rational(max([s.real() for s in section]), 0.1)
                xmin=Util.simple_rational(min([s.real() for s in section]), 0.1)
                ymax=Util.simple_rational(max([s.imag() for s in section]), 0.1)
                clockwise = Util.is_clockwise([section[i] if i!=self.npoints else 2*xmin-xmax+ymax*I/5 for i in cycle])

            
            # logger.info("Clockwise cycle" if clockwise else "Counterclockwise cycle")

            beforeLoop = True
            beforeLink = True
            beforeEdges1 = []
            beforeEdges2 = []
            afterEdges = []
            link=min(cycle, key=lambda v:ga.distance(v, self.npoints))

            for j in range(len(cycle)-1):
                e2 = (cycle[j], cycle[j+1])
                if link==cycle[j]:
                    beforeLink = False
                if e2[0]>e2[1]:
                    e2=(e2[1],e2[0])
                if self.normalize_edge(e2) == ea:
                    beforeLoop = False
                    gx.add_edge(ea)
                else:
                    if beforeLink:
                        beforeEdges1+=[e2]
                    elif beforeLoop:
                        beforeEdges2+=[e2]
                    else:
                        afterEdges+=[e2]
            e = self.normalize_edge(e)

            inside = []
            outside = []
            for j in range(len(cycle)):
                before, current, after = [cycle[l%len(cycle)] for l in [j-1, j, j+1]] 
                neighbours = self.neighbours(ga, current, section)
                neighbours = neighbours[neighbours.index(before):] + neighbours[:neighbours.index(before)]
                if clockwise:
                    inside += [[current, v] for v in neighbours[neighbours.index(after)+1:]]
                    outside += [[current, v] for v in neighbours[1:neighbours.index(after)]]
                else:
                    inside += [[current, v] for v in neighbours[1:neighbours.index(after)]]
                    outside += [[current, v] for v in neighbours[neighbours.index(after)+1:]]
            # then add all the edges rooted at points of inside
            inside2 = []
            for ei in inside:
                gtempx = copy(ga)
                gtempx.delete_vertex(ei[0])
                inside2 += self.ordered_edges(gtempx.subgraph(gtempx.connected_component_containing_vertex(ei[1])))
                inside2 += [self.normalize_edge(ei)]

            outside2 = []
            for ei in outside:
                gtempx = copy(ga)
                gtempx.delete_vertex(ei[0])
                outside2 += self.ordered_edges(gtempx.subgraph(gtempx.connected_component_containing_vertex(ei[1])))
                outside2 += [self.normalize_edge(ei)]

            basepointisinside = self.hasbasepoint and self.npoints in flatten(inside2) and self.npoints not in cycle
            otherside = outside2 if basepointisinside else inside2
            sameside = inside2 if basepointisinside else outside2

            res = []
            ini = self.ordered_edges(g1)
            fin = self.ordered_edges(gx)
            xr= self.xs[fin.index(ea)]

            for e2 in ini:
                if (clockwise and not basepointisinside) or (not clockwise and basepointisinside):
                    if e2 in beforeEdges1:
                        res+=[self.xs[fin.index(e2)]*xr]
                    elif e2 in beforeEdges2:
                        res+=[xr**-1*self.xs[fin.index(e2)]]
                    elif e2 in afterEdges:
                        res+=[self.xs[fin.index(e2)]**-1*xr]
                    elif e2 == e:
                        res+=[xr]
                    elif e2 in otherside:
                        res+=[xr**-1*self.xs[fin.index(e2)]*xr]
                    elif e2 in sameside:
                        res+=[self.xs[fin.index(e2)]]
                    else:
                        logger.warning("unrecognized edge")
                else:
                    if e2 in beforeEdges1:
                        res+=[xr*self.xs[fin.index(e2)]]
                    elif e2 in beforeEdges2:
                        res+=[self.xs[fin.index(e2)]*xr**-1]
                    elif e2 in afterEdges:
                        res+=[xr*self.xs[fin.index(e2)]**-1]
                    elif e2 == e:
                        res+=[xr]
                    elif e2 in otherside:
                        res+=[xr*self.xs[fin.index(e2)]*xr**-1]
                    elif e2 in sameside:
                        res+=[self.xs[fin.index(e2)]]
                    else:
                        logger.warning("unrecognized edge")

            iso = self.freeGroup.hom([self.freeGroup.hom(res)(iso(x)) for x in self.xs])
            g1=gx
        # logger.info("Ended with graph with edges (%d, %d), (%d, %d), (%d, %d)."% tuple(flatten([[e[0], e[1]] for e in g1.edges()])))
        assert g1==g2, "Did not recover correct graph"
        return iso

    def normalize_edge(self, e):
        return (e[0], e[1]) if e[0]<=e[1] else (e[1], e[0])

    def ordered_edges(self, gr):
        res = [self.normalize_edge(e) for e in gr.edges()]
        res.sort()
        return res
    
    def neighbours(self, g, v, section):
        neighbours = g.neighbors(v)
        if self.hasbasepoint:
            section2 = section + [self.basepoint]
        else:
            xmax=Util.simple_rational(max([s.real() for s in section]), 0.1)
            xmin=Util.simple_rational(min([s.real() for s in section]), 0.1)
            ymax=Util.simple_rational(max([s.imag() for s in section]), 0.1)
            section2 = section + [2*xmin-xmax+ymax*I/5]
        neighbours.sort(key=lambda v2:-arg(section2[v2] - section2[v]))
        return neighbours


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import sage.all

from .numperiods.family import Family
from .numperiods.cohomology import Cohomology
from .numperiods.integerRelations import IntegerRelations
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.qqbar import AlgebraicField
from sage.rings.qqbar import QQbar
from sage.functions.other import factorial
from sage.matrix.constructor import matrix
from sage.arith.misc import xgcd
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix
from sage.arith.functions import lcm
from sage.quadratic_forms.quadratic_form import QuadraticForm
from sage.misc.flatten import flatten

from ore_algebra.analytic.differential_operator import DifferentialOperator


from sage.misc.prandom import randint

from .voronoi import FundamentalGroupVoronoi
from .integrator_simultaneous import IntegratorSimultaneous
from .integrator import Integrator
from .util import Util
from .context import Context
from .exceptionalDivisorComputer import ExceptionalDivisorComputer
from .delaunayDual import FundamentalGroupDelaunayDual
from .monodromyRepresentationGeneric import MonodromyRepresentationGeneric
from .monodromyRepresentationSurface import MonodromyRepresentationSurface

import logging
import time

logger = logging.getLogger(__name__)


class Hypersurface(object):
    def __init__(self, P, basepoint=None, fibration=None, compute_fundamental_group=True, **kwds):
        """P, a homogeneous polynomial defining a smooth hypersurface X in P^{n+1}.

        This class aims at computing an effective basis of the homology group H_n(X), 
        given as lifts of paths through a Lefschetz fibration.
        """
        
        self.ctx = Context(**kwds)
        
        # assert P.is_homogeneous(), "nonhomogeneous defining polynomial"
        
        self._P = P
        if fibration != None:
            assert matrix(fibration).rank() == len(fibration), "collection of hyperplanes is not generic"
            assert len(fibration) == self.dim+1, "fibration does not have the correct number of hyperplanes"
            self._fibration = fibration
        if self.dim>=1 and not self.ctx.debug:
            fg = self.fundamental_group # this allows reordering the critical points straight away and prevents shenanigans. There should be a better way to do this

        if basepoint!= None: # it is useful to be able to specify the basepoint to avoid being stuck in arithmetic computations if critical values have very large modulus
            assert basepoint not in self.critical_values, "basepoint is not regular"
            self._basepoint = basepoint
    
    def __str__(self):
        s = "Hypersurface of dimension " + str(self.dim)+" and degree " + str(self.degree)
        return s
    
    def __repr__(self):
        s = "Hypersurface of dimension " + str(self.dim)+" and degree " + str(self.degree)
        return s
    
    @property
    def intersection_product_modification(self):
        """The intersection matrix of the modification of the hypersurface"""
        if not hasattr(self,'_intersection_product_modification'):
            assert self.dim!=0, "no modification in dimension 0"
            self._intersection_product_modification = self.monodromy_representation.intersection_product
        return self._intersection_product_modification

    @property
    def monodromy_representation(self):
        """The monodromy representation associated to the modification of the hypersurface"""
        if not hasattr(self,'_monodromy_representation'):
            assert self.dim!=0, "no monodromy_representation in dimension 0"
            if self.dim == 2:
                monodromy_representation = MonodromyRepresentationSurface(self.monodromy_matrices, self.fibre.intersection_product)
            else:
                monodromy_representation = MonodromyRepresentationGeneric(self.monodromy_matrices, self.fibre.intersection_product)
                monodromy_representation._add = 0 if self.dim%2==1 else 2
            self._monodromy_representation = monodromy_representation
        return self._monodromy_representation
    
    @property
    def intersection_product(self):
        """The intersection matrix of the hypersurface"""
        if not hasattr(self,'_intersection_product'):
            if self.dim==0:
                self._intersection_product = identity_matrix(self.degree)
            elif self.dim==1:
                self._intersection_product = self.intersection_product_modification
            else:
                homology = matrix(self.homology)
                IP = homology * self.intersection_product_modification * homology.transpose()
                assert IP.det() in [1,-1], "intersection product is not unitary"
                self._intersection_product = IP
        return self._intersection_product

    @property
    def homology(self):
        """An embedding of the homology of the hypersurface in the homology of the modification."""
        if not hasattr(self,'_homology'):
            if self.dim<2:
                self._homology = identity_matrix(len(self.extensions)).rows()
            else:
                exceptional_divisors = matrix(self.exceptional_divisors).transpose()
                product_with_exdiv = self.intersection_product_modification * exceptional_divisors
                product_with_exdiv = product_with_exdiv.change_ring(ZZ)
                self._homology = product_with_exdiv.kernel().basis()
        return self._homology

    @property
    def period_matrix_modification(self):
        """The period matrix of the modification of the hypersurface"""
        if not hasattr(self, '_period_matrix_modification'):
            integrated_thimbles = self.integrated_thimbles
            add  = [vector([0]*len(self.monodromy_representation.thimbles))] * len(flatten(self.monodromy_representation.components_of_singular_fibres))
            add += [vector([0]*len(self.monodromy_representation.thimbles))] * 2 if self.dim%2 ==0 else []
            homology_mat = matrix(self.monodromy_representation.extensions + add).transpose()
            primary_lattice = self.monodromy_representation.primary_lattice
            self._period_matrix_modification =  integrated_thimbles * homology_mat * primary_lattice.inverse()
        return self._period_matrix_modification

    @property
    def period_matrix(self):
        """The period matrix of the hypersurface"""
        if not hasattr(self, '_period_matrix'):
            if self.dim==0:
                R = self.P.parent()
                affineR = PolynomialRing(QQbar, 'X')
                affineProjection = R.hom([affineR.gens()[0],1], affineR)
                period_matrix = matrix([self._residue_form(affineProjection(b), affineProjection(self.P), (b.degree()+len(R.gens()))//self.P.degree(), self.extensions) for b in self.cohomology]).change_ring(self.ctx.CBF)
                period_matrix = block_matrix([[period_matrix],[matrix([[1]*self.degree])]])
                self._period_matrix=period_matrix
            elif self.dim%2 ==1:
                self._period_matrix = self.period_matrix_modification * matrix(self.homology).transpose()
            else:
                self._period_matrix = block_matrix([[self.period_matrix_modification*matrix(self.homology).transpose()], [matrix(self.intersection_product*self.lift_modification(self.fibre_class))]])
        return self._period_matrix
    
    @property
    def holomorphic_period_matrix_modification(self):
        """The holomorphic period matrix of the modification of the hypersurface"""
        if not hasattr(self, '_holomorphic_period_matrix_modification'):
            if self.dim==0:
                self._holomorphic_period_matrix_modification = self.period_matrix
            else:
                integrated_thimbles_holomorphic = self.integrated_thimbles_holomorphic
                add = [vector([0]*len(self.monodromy_representation.thimbles))] * len(flatten(self.monodromy_representation.components_of_singular_fibres))
                add += [vector([0]*len(self.monodromy_representation.thimbles))] * 2 if self.dim%2 ==0 else []
                homology_mat = matrix(self.monodromy_representation.extensions + add).transpose()
                primary_lattice = self.monodromy_representation.primary_lattice
                self._holomorphic_period_matrix_modification =  integrated_thimbles_holomorphic * homology_mat * primary_lattice.inverse()
        return self._holomorphic_period_matrix_modification

    @property
    def holomorphic_period_matrix(self):
        """The holomorphic period matrix of the hypersurface"""
        if not hasattr(self, '_holomorphic_period_matrix'):
            if self.dim==0:
                self._holomorphic_period_matrix = self.period_matrix
            else:
                periods_modification = self.holomorphic_period_matrix_modification
                homology = matrix(self.homology).transpose()
                self._holomorphic_period_matrix = periods_modification * homology
        return self._holomorphic_period_matrix

    @property
    def holomorphic_forms(self):
        """The holomorphic cohomology classes."""
        if not hasattr(self, "_holomorphic_forms"):
            mindeg = min([m.degree() for m in self.cohomology])
            self._holomorphic_forms = [m for m in self.cohomology if m.degree()==mindeg]
        return self._holomorphic_forms


    @property
    def P(self):
        """The defining equation of the hypersurface."""
        return self._P

    @property
    def degree(self):
        if not hasattr(self,'_degree'):
            self._degree = self.P.degree()
        return self._degree
    
    @property
    def dim(self):
        if not hasattr(self,'_dim'):
            self._dim = len(self.P.parent().gens())-2
        return self._dim

    @property
    def cohomology(self):
        if not hasattr(self,'_cohomology'):
            self._cohomology = Cohomology(self.P).basis()
        return self._cohomology
    
    
    @property
    def family(self):
        if not hasattr(self,'_family'):
            RtoS = self._RtoS()
            self._family = Family(RtoS(self.P))
        return self._family
    

    @property
    def fibration(self):
        if not hasattr(self,'_fibration'): #TODO try to reduce variance of distance between critical points(?)
            rank = self.dim+1 if self.ctx.long_fibration else 2
            
            r=0
            fibration = []
            for r in range(rank):
                while True:
                    v = vector([randint(-10,10) for i in range(self.dim+2)])
                    if v not in matrix(fibration).image():
                        fibration += [v]
                        break
            self._fibration = fibration
        return self._fibration

    @property
    def critical_values(self):
        if not hasattr(self,'_critical_values'):
            R = self.P.parent()
            _vars = [v for v in R.gens()]
            forms=[v.dot_product(vector(_vars)) for v in self.fibration[:2]]
            f=forms[0]/forms[1]
            S = PolynomialRing(QQ, _vars+['k','t'])
            k,t= S.gens()[-2:]
            eqs = [
                self.P, 
                forms[1]-1, 
                t*forms[1]-forms[0]
            ] + [(f.derivative(var).numerator()*k-self.P.derivative(var)*f.derivative(var).denominator()) for var in _vars]

            ideal = S.ideal(eqs).elimination_ideal(S.gens()[:-1])
            Qt = PolynomialRing(QQ, 't')

            roots_with_multiplicity = Qt(ideal.groebner_basis()[0]).roots(AlgebraicField())
            if not self.ctx.debug and not self.ctx.singular:
                for e in roots_with_multiplicity:
                    assert e[1]==1, "double critical values, fibration is not Lefschetz"
            self._critical_values=[e[0] for e in roots_with_multiplicity]
        return self._critical_values
    
    @property
    def monodromy_matrices(self):
        assert self.dim!=0, "Cannot compute monodromy matrices in dimension 0"
        if not hasattr(self, '_monodromy_matrices'):
            i=0
            transition_matrices = self.transition_matrices_monodromy

            n = len(self.fibre.homology) 
            
            cohomology_fibre_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fibre.cohomology], self.basepoint)
            initial_conditions = cohomology_fibre_to_family.inverse()

            cohomology_monodromies = [initial_conditions.inverse() * M * initial_conditions for M in transition_matrices]
            if self.dim%2==1:
                cohomology_monodromies = [block_diagonal_matrix([M, identity_matrix(1)]) for M in cohomology_monodromies]

            Ms = [(self.fibre.period_matrix.inverse() * M * self.fibre.period_matrix) for M in cohomology_monodromies]
            try:
                Ms = [M.change_ring(ZZ) for M in Ms]
            except:
                if self.ctx.debug:
                    logger.info("Monodromy is not integral")
                else:
                    raise Exception("Monodromy is not integral")
                
            if not self.ctx.singular and not self.ctx.debug:
                for M in Ms:
                    assert (M-1).rank()==1, "If M is a monodromy matrix around a single critical point, M-1 should have rank 1"
            
            self._monodromy_matrices = Ms
        return self._monodromy_matrices
    
    @property
    def fibre(self):
        assert self.dim!=0, "Variety of dimension 0 does not have a fibre"
        if not hasattr(self,'_fibre'):
            RtoS = self._RtoS()
            evaluate_at_basepoint = RtoS.codomain().hom([self.basepoint], RtoS.codomain().base_ring())
            P = evaluate_at_basepoint(RtoS(self.P))
            fibration  = self._restrict_fibration() if self.ctx.long_fibration else None
            self._fibre = Hypersurface(P, 
                                       fibration=fibration, 
                                       method=self.ctx.method, 
                                       nbits=self.ctx.nbits, 
                                       long_fibration=self.ctx.long_fibration, 
                                       depth=self.ctx.depth+1,
                                       simultaneous_integration=True
                                       )

        return self._fibre

    @property
    def thimbles(self):
        return self.monodromy_representation.thimbles

    @property
    def permuting_cycles(self):
        return self.monodromy_representation.permuting_cycles

    @property
    def vanishing_cycles(self):
        return flatten(self.monodromy_representation.vanishing_cycles_desingularisation)

    @property
    def extensions(self):
        if not hasattr(self, '_extensions'):
            if self.dim==0:
                R = self.P.parent()
                affineR = PolynomialRing(QQbar, 'X')
                affineProjection = R.hom([affineR.gens()[0],1], affineR)
                self._extensions = [e[0] for e in affineProjection(self.P).roots()]
                return self._extensions
            self._extensions = self.monodromy_representation.extensions
        return self._extensions

    @property
    def thimble_extensions(self):
        """Returns a list of elements of the form `[boundary, extension]` such that `extension` is the extension of a thimble of the fibre of the fibre along the loop around infinity. 
        `boundary` is the boundary of the extended thimble. 
        The list consisting of the `boundary`s forms a basis of the image of the boundary map.
        `extension` if defined only up to an infinity loop."""
        if not hasattr(self, '_thimble_extensions'):
            if self.dim<2:
                self._thimble_extensions = []
            else:
                distinct_vanishing_cycles = []
                for i, v in enumerate(self.fibre.vanishing_cycles):
                    if v not in matrix([self.fibre.vanishing_cycles[j] for j in distinct_vanishing_cycles]).image():
                        distinct_vanishing_cycles+=[i]
                        
                chains = [self.fibre.vanishing_cycles[i] for i in distinct_vanishing_cycles]

                thimble_extensions = zero_matrix(len(distinct_vanishing_cycles),len(self.thimbles))
                for j in distinct_vanishing_cycles:
                    u=vector([0]*len(self.fibre.thimbles))
                    u[j]=1
                    for i, M in enumerate(self.thimble_monodromy):
                        v = (M-1)*u
                        v = self.fibre.lift(v)
                        u = M*u
                        c = v/self.vanishing_cycles[i]
                        thimble_extensions[j,i] = c
                thimble_extensions = thimble_extensions.rows()
                self._thimble_extensions = list(zip(chains, thimble_extensions))
        return self._thimble_extensions

    @property
    def thimble_monodromy(self):
        if not hasattr(self, "_thimble_monodromy"):
            logger.info("[%d] Computing thimble monodromy with braids, this may take a while."% self.dim)
            begin = time.time()
            self._EDC = ExceptionalDivisorComputer(self)
            self._thimble_monodromy = self._EDC.thimble_monodromy
            end = time.time()
            duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
            logger.info("Thimble monodromy computed in %s.", duration_str)
        return self._thimble_monodromy
    
    @property
    def invariant(self): # TODO : in dim >0 this is just the fibre. 
        if not hasattr(self, '_invariant'):
            self._invariant = vector(Util.find_complement(matrix([chain for chain, _ in self.thimble_extensions])))
        return self._invariant

    @property
    def exceptional_divisors(self):
        """Returns the coordinates of the exceptional divisors in the basis of homology of the modification."""
        if not hasattr(self, '_exceptional_divisors'):
            if self.dim==2 and self.degree in [3,4]: # this is specific for K3 surfaces, while waiting for more robust/efficient methods arrive for the general case
                if self.degree ==4:
                    NS = IntegerRelations(self.holomorphic_period_matrix_modification.transpose()).basis
                if self.degree==3:
                    NS = IntegerRelations(self.period_matrix_modification.transpose()).basis
                section = self.section
                fibre = self.fibre_class
                others = (NS * self.intersection_product_modification * matrix([section, fibre]).transpose()).kernel().basis_matrix()
                short_vectors = QuadraticForm(-2 * others * NS * self.intersection_product_modification * NS.transpose() * others.transpose()).short_vector_list_up_to_length(3)[2]
                short_vectors = [v * others * NS for v in short_vectors]
                exp_divs = [v+section+fibre for v in short_vectors]
                chosen=[]
                expected_number = self.degree-1 # this is a temporary workaround
                while len(exp_divs)!=0:
                    if len([v for v in exp_divs if v * self.intersection_product_modification*exp_divs[0]==0])>0 or len(chosen) == expected_number-1:
                        chosen += [exp_divs[0]]
                        exp_divs = [v for v in exp_divs if v * self.intersection_product_modification*chosen[-1]==0]
                    else:
                        exp_divs = exp_divs[1:]
                self._exceptional_divisors = chosen + [self.section]
            else:
                if self.dim%2 ==1:
                    exceptional_divisors = [self.lift(extension) for _, extension in self.thimble_extensions]
                    chains = matrix([chain for chain, _ in self.thimble_extensions])
                else:
                    exceptional_divisors = []
                    for chain, extension in self.thimble_extensions:
                        exceptional_divisor = self.lift(extension)
                        exceptional_divisor -= self.fibre_class * (chain*self.fibre.fibre.intersection_product*self.invariant)
                        exceptional_divisors += [exceptional_divisor]
                    exceptional_divisors += [self.section]
                    chains = matrix([chain for chain, _ in self.thimble_extensions] + [self.invariant])
                self._exceptional_divisors = (chains**(-1)*matrix(exceptional_divisors)).rows()
        return self._exceptional_divisors

    @property
    def _restriction_variable(self):
        if not hasattr(self, "__restriction_variable"):
            for i in range(self.dim+2):
                if self.fibration[0][i] !=0:
                    break
            self.__restriction_variable = i # type: ignore
        return self.__restriction_variable

    def _RtoS(self):
        R = self.P.parent()

        a = self.fibration[1][self._restriction_variable]
        b = self.fibration[0][self._restriction_variable]

        _vars = [v for v in R.gens()]
        S = PolynomialRing(PolynomialRing(QQ, [_vars[i] for i in range(len(_vars)) if i != self._restriction_variable]), 't')
        t = S.gens()[0]

        l = vector([c for i, c in enumerate(_vars) if i != self._restriction_variable])*vector([c for i,c in enumerate(self.fibration[0]) if i != self._restriction_variable])
        m = vector([c for i, c in enumerate(_vars) if i != self._restriction_variable])*vector([c for i,c in enumerate(self.fibration[1]) if i != self._restriction_variable])

        form = (-S(l)+t*S(m))
        denom = b-a*t

        RtoS = R.hom([denom*S(_vars[i]) if i != self._restriction_variable else form for i in range(len(_vars))], S)
        return RtoS
    
    def _restrict_fibration(self):
        a = self.fibration[1][self._restriction_variable]
        b = self.fibration[0][self._restriction_variable]
        
        l = vector([c for i, c in enumerate(self.fibration[0]) if i != self._restriction_variable])
        m = vector([c for i, c in enumerate(self.fibration[1]) if i != self._restriction_variable])

        form = -l+self.basepoint*m
        
        hyperplanes = []
        for hyperplane in self.fibration[1:]:
            cs = (b-a*self.basepoint) * vector([c for i, c in enumerate(hyperplane) if i != self._restriction_variable])
            cz = hyperplane[self._restriction_variable] * form
            hyperplanes += [cs+cz]
        return hyperplanes

    def _restrict_form(self, A):
        """ Given a form A, returns the form A_t such that A/P^k w_n = A_t/P_t^k w_{n-1}dt
        """
        assert self.dim !=0, "cannot restrict form of a dimension 0 variety"

        RtoS = self._RtoS()


        R = self.P.parent()

        a = self.fibration[1][self._restriction_variable]
        b = self.fibration[0][self._restriction_variable]

        _vars = [v for v in R.gens()]
        S = PolynomialRing(PolynomialRing(QQ, [_vars[i] for i in range(len(_vars)) if i != self._restriction_variable]), 't')
        t = S.gens()[0]

        l = vector([_vars[i] for i in range(len(_vars)) if i != self._restriction_variable])*vector([self.fibration[0][i] for i in range(len(_vars)) if i != self._restriction_variable])
        m = vector([_vars[i] for i in range(len(_vars)) if i != self._restriction_variable])*vector([self.fibration[1][i] for i in range(len(_vars)) if i != self._restriction_variable])

        form = (-S(l)+t*S(m))
        denom = b-a*t

        dt = (-a * S(l) +b* S(m))*denom**(self.dim)
        return RtoS(A)*dt

    @property
    def transition_matrices(self):
        if not hasattr(self, '_transition_matrices'):
            if hasattr(self, '_transition_matrices_holomorphic') and len(self.holomorphic_forms) == len(self.cohomology):
                self._transition_matrices = self.transition_matrices_holomorphic
                return self._transition_matrices

            if hasattr(self, '_transition_matrices_holomorphic'):
                rat_coefs = self.family.coordinates([self._restrict_form(w) for w in self.cohomology if w not in self.holomorphic_forms])
            else:
                rat_coefs = self.family.coordinates([self._restrict_form(w) for w in self.cohomology])

            logger.info("[%d] Computing transition matrices of all forms (%d rational vector(s))."% (self.dim, rat_coefs[0].nrows()))
            if rat_coefs[0].nrows() > self.ctx.cutoff_simultaneous_integration:
                transition_matrices = self._compute_transition_matrices_simultaneous(rat_coefs)
            else:
                indices = [i for i in range(len(self.cohomology))]
                transition_matrices = self._compute_transition_matrices_sequential(rat_coefs, indices)
            self._transition_matrices = transition_matrices
        return self._transition_matrices

    @property
    def transition_matrices_holomorphic(self):
        if not hasattr(self, '_transition_matrices_holomorphic'):
            if hasattr(self, '_transition_matrices') or self.ctx.simultaneous_integration:
                r = self.fibre.period_matrix.nrows()
                r = len(self.fibre.cohomology)
                R = len(self.cohomology)
                indices = [self.cohomology.index(w) for w in self.holomorphic_forms]
                indices += [R+i for i in range(r)]
                transition_matrices = [M.matrix_from_rows_and_columns(indices, indices) for M in self.transition_matrices]
            else:
                rat_coefs = self.family.coordinates([self._restrict_form(w) for w in self.holomorphic_forms])
                logger.info("[%d] Computing transition matrices for holomorphic forms (%d rational vector(s))."% (self.dim, rat_coefs[0].nrows()))
                if rat_coefs[0].nrows() > self.ctx.cutoff_simultaneous_integration:
                    transition_matrices = self._compute_transition_matrices_simultaneous(rat_coefs)
                else:
                    indices = [i for i in range(len(self.holomorphic_forms))]
                    transition_matrices = self._compute_transition_matrices_sequential(rat_coefs, indices)
            self._transition_matrices_holomorphic = transition_matrices
        return self._transition_matrices_holomorphic

    @property
    def transition_matrices_monodromy(self): # this function might very well be buggy, but it's not called very often.
        if not hasattr(self, '_transition_matrices_monodromy'):
            if hasattr(self, '_transition_matrices') or self.ctx.simultaneous_integration:
                r = self.fibre.period_matrix.nrows()
                R = len(self.cohomology)
                transition_matrices = [M.submatrix(R,R) for M in self.transition_matrices]
            elif hasattr(self, '_transition_matrices_holomorphic'):
                r = self.fibre.period_matrix.nrows()
                R = len(self.holomorphic_forms)
                transition_matrices = [M.submatrix(R,R) for M in self.transition_matrices_holomorphic]
            else:
                indices = [0]
                rat_coefs = self.family.coordinates([self._restrict_form(self.cohomology[0])])
                logger.info("[%d] Computing transition matrices for monodromy (%d rational vector(s))."% (self.dim, rat_coefs[0].nrows()))
                transition_matrices = self._compute_transition_matrices_sequential(rat_coefs, indices)
                transition_matrices = [M.submatrix(1,1) for M in transition_matrices]
            self._transition_matrices_monodromy = transition_matrices
        return self._transition_matrices_monodromy

    def _compute_transition_matrices_simultaneous(self, rat_coefs):
        logger.info("[%d] Computing Gauss-Manin connection."% (self.dim))
        begin = time.time()
        gaussmanin = self.family.gaussmanin()
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("[%d] Gauss-Manin connection computed in %s."% (self.dim, duration_str))

        logger.info("[%d] Computing numerical transition matrices for %d integrals (%d edges total)."% (self.dim, rat_coefs[0].nrows(), len(self.fundamental_group.edges)))
        begin = time.time()
        integrator = IntegratorSimultaneous(self.fundamental_group, rat_coefs, gaussmanin, self.ctx.nbits)
        transition_matrices = integrator.transition_matrices
        if hasattr(self, '_transition_matrices_holomorphic'):
            Rholo = len(self.holomorphic_forms)
            R = len(self.cohomology) - Rholo
            r = len(self.fibre.cohomology)
            intold = [M.submatrix(0,Rholo,Rholo,r) for M in self.transition_matrices_holomorphic]
            intnew = [M.submatrix(0,R,R,r) for M in transition_matrices]
            GM = [M.submatrix(R,R) for M in transition_matrices]
            transition_matrices = [block_matrix([[1,0, a],[0,1, b], [0,0,c]]) for a,b,c in zip(intold, intnew, GM)]
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("[%d] Integration finished -- total time: %s."% (self.dim, duration_str))
        return transition_matrices

    def _compute_transition_matrices_sequential(self, rat_coefs, indices):
        R, denom = rat_coefs
        res = None
        j=0
        logger.info("[%d] Computing Picard-Fuchs equations of %d form(s) in dimension %d"% (self.dim, R.nrows(), self.dim))
        for i, v in zip(indices, R.rows()):
            L = self.picard_fuchs_equation(v/denom)
            L = L * L.parent().gens()[0]
            logger.info("[%d] Integrating operator [%d/%d] with order %d and degree %d."% (self.dim, j+1, R.nrows(), L.order(), L.degree()))
            integrated = self.integrate(L)
            derivatives_at_basepoint = self.derivatives_values_at_basepoint(i)
            integration_correction = diagonal_matrix([1/ZZ(factorial(k)) for k in range(L.order())])
            initial_conditions = ( integration_correction * derivatives_at_basepoint )
            if res == None:
                res = []
                for M in integrated:
                    GM = initial_conditions.submatrix(1,0).inverse() * M.submatrix(1,1) * initial_conditions.submatrix(1,0)
                    intgr = (M * initial_conditions).submatrix(0,0,1)
                    res += [block_matrix([[identity_matrix(1), intgr],[0, GM]])]
            else:
                r = L.order()-1
                intold = [M.submatrix(0,j,j,r) for M in res]
                intnew = [(M * initial_conditions).submatrix(0,0,1) for M in integrated]
                GM = [M.submatrix(j,j) for M in res]
                res = [block_matrix([[1,0, a],[0,1, b], [0,0,c]]) for a,b,c in zip(intold, intnew, GM)]
            j+=1
        return res

    def picard_fuchs_equation(self, v):
        denom = lcm([r.denominator() for r in v if r!=0])
        numerators = denom * v
        L = self.family.picard_fuchs_equation(numerators) * denom
        L = DifferentialOperator(L)
        return L

    def integrate(self, L):
        logger.info("[%d] Computing numerical transition matrices of operator of order %d and degree %d (%d edges total)."% (self.dim, L.order(), L.degree(), len(self.fundamental_group.edges)))
        begin = time.time()
        integrator = Integrator(self.fundamental_group, L, self.ctx.nbits)
        transition_matrices = integrator.transition_matrices
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("[%d] Integration finished -- total time: %s."% (self.dim, duration_str))
        return transition_matrices

    @property
    def integrated_thimbles(self):
        if not hasattr(self, '_integrated_thimbles'):
            s=len(self.fibre.homology)
            transition_matrices = self.transition_matrices
            R=len(self.cohomology)
            permuting_cycles = self.permuting_cycles

            cohomology_fibre_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fibre.cohomology], self.basepoint)
            initial_conditions = cohomology_fibre_to_family.inverse()
            
            pM = self.fibre.period_matrix
            if self.dim%2==1:
                pM = pM.submatrix(0,0,s-1)
            expand = zero_matrix(R,s-1 if self.dim%2==1 else s).stack(identity_matrix(s-1 if self.dim%2==1 else s))
            integrated_thimbles = []
            for tM, pcs in zip(transition_matrices, permuting_cycles):
                for pc in pcs:
                    integrated_thimbles += [(tM * expand * initial_conditions * pM * pc)[:R]]
            self._integrated_thimbles = matrix(integrated_thimbles).transpose()
        return self._integrated_thimbles
    
    @property
    def integrated_thimbles_holomorphic(self):
        if not hasattr(self, '_integrated_thimbles_holomorphic'):
            s=len(self.fibre.homology)
            transition_matrices = self.transition_matrices_holomorphic
            R=len(self.holomorphic_forms)
            permuting_cycles = self.permuting_cycles

            cohomology_fibre_to_family = self.family._coordinates([self.family.pol.parent()(w) for w in self.fibre.cohomology], self.basepoint)
            initial_conditions = cohomology_fibre_to_family.inverse()
            
            pM = self.fibre.period_matrix
            if self.dim%2==1:
                pM = pM.submatrix(0,0,s-1)
            expand = zero_matrix(R,s-1 if self.dim%2==1 else s).stack(identity_matrix(s-1 if self.dim%2==1 else s))
            integrated_thimbles = []
            for tM, pcs in zip(transition_matrices, permuting_cycles):
                for pc in pcs:
                    integrated_thimbles += [(tM * expand * initial_conditions * pM * pc)[:R]]
            self._integrated_thimbles_holomorphic = matrix(integrated_thimbles).transpose()
        return self._integrated_thimbles_holomorphic
    
    # Integration methods

    def derivatives_coordinates(self, i):
        if not self._coordinatesQ[i]:
            s=len(self.fibre.homology)
            RtoS = self._RtoS()
            w = self.cohomology[i]
            wt = self._restrict_form(w)
            derivatives = [RtoS(0), wt]
            for k in range(s-1 if self.dim%2==0 else s-2):
                derivatives += [self._derivative(derivatives[-1], RtoS(self.P))] 
            self._coordinates[i] = self.family.coordinates(derivatives)
            self._coordinatesQ[i] = True
        return self._coordinates[i]

    def _residue_form(self, A, P, k, alphas): 
        """ returns the formal residue of A/P^k at alpha for alpha in alphas """
        G,U,V = xgcd(P, P.derivative())
        assert G==1, "P is not squarefree"
        if k==1:
            return [V(alpha)*A(alpha) for alpha in alphas]
        return self._residue_form(A*U/(k-1)+(A*V).derivative()/(k-1)**2, P, k-1, alphas)
    
    def derivatives_values_at_basepoint(self, i):
        RtoS = self._RtoS()
        s=len(self.fibre.cohomology)
        w = self.cohomology[i]
        wt = self._restrict_form(w)
        derivatives = [RtoS(0), wt]
        for k in range(s-1):
            derivatives += [self._derivative(derivatives[-1], RtoS(self.P))] 
        return self.family._coordinates(derivatives, self.basepoint)

    @classmethod
    def _derivative(self, A, P): 
        """computes the numerator of the derivative of A/P^k"""
        return A.derivative() - A*P.derivative()         

    @property
    def fundamental_group(self):
        assert self.dim>0, "Dimension 0 vartiety has no fibration"
        if not hasattr(self,'_fundamental_group'):
            logger.info("[%d] Computing fundamental group with %d critical values."% (self.dim, len(self.critical_values)))
            begin = time.time()
            if self.ctx.method == 'voronoi':# access future delaunay implem here
                fundamental_group = FundamentalGroupVoronoi(self.critical_values, self.basepoint)
            elif self.ctx.method == 'delaunay_dual':
                fundamental_group = FundamentalGroupDelaunayDual(self.critical_values, self.basepoint)
            else:
                fundamental_group = FundamentalGroupVoronoi(self.critical_values, self.basepoint)
            fundamental_group.sort_loops()

            end = time.time()
            duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
            logger.info("[%d] Fundamental group computed in %s."% (self.dim, duration_str))

            self._critical_values = fundamental_group.points[1:]
            self._fundamental_group = fundamental_group
        return self._fundamental_group

    @property
    def paths(self):
        assert self.dim>0, "Dimension 0 vartiety has no fibration"
        if not hasattr(self,'_paths'):
            paths = []
            for path in self.fundamental_group.pointed_loops:
                paths += [[self.fundamental_group.vertices[v] for v in path]]
            self._paths = paths
        return self._paths

    @property
    def basepoint(self):
        assert self.dim>0, "Dimension 0 vartiety has no fibration"
        if  not hasattr(self, '_basepoint'):
            if self.ctx.long_fibration and self.ctx.depth>0:
                self._basepoint=ZZ(0)
            else:
                shift = 1
                reals = [self.ctx.CF(c).real() for c in self.critical_values]
                xmin, xmax = min(reals), max(reals)
                self._basepoint = Util.simple_rational(xmin - (xmax-xmin)*shift, (xmax-xmin)/10)
        return self._basepoint

    def lift(self, v):
        return self.monodromy_representation.lift(v)
    
    def lift_modification(self, v):
        """Given a vector v, return the orthogonal projection of the homology of the modification on the homology of the hypersurface"""
        return vector(list(matrix(self.homology + self.exceptional_divisors).solve_left(v))[:len(self.homology)])
    
    @property
    def hyperplane_class(self):
        assert self.dim %2 ==0, "no hyperplane class in odd dimensions"
        return matrix(self.homology).solve_left(self.fibre_class + sum(self.exceptional_divisors)) # todo in higher dimensions (>=4) we want the orthogonal projection of self.fibre_class
    
    @property
    def fibre_class(self):
        return self.monodromy_representation.fibre_class
    @property
    def section(self):
        return self.monodromy_representation.section


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from .numperiods.family import Family
from .numperiods.integerRelations import IntegerRelations
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.qqbar import QQbar
from sage.functions.other import factorial
from sage.functions.other import floor
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix

from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.schemes.toric.weierstrass import WeierstrassForm
from sage.misc.flatten import flatten

from sage.modules.free_quadratic_module_integer_symmetric import IntegralLattice

from .voronoi import FundamentalGroupVoronoi
from .integrator import Integrator
from .util import Util
from .context import Context
from .hypersurface import Hypersurface
from .monodromyRepresentationEllipticSurface import MonodromyRepresentationEllipticSurface

import logging
import time

logger = logging.getLogger(__name__)


class AllowableLoop(object):
    def __init__(self, surface, loop, **kwds):
        """surface, an elliptic surface, loop an allowable loop.

        This class aims at computing an effective basis of the homology group H_n(X), 
        given as lifts of paths through a Lefschetz fibration.
        """
        
        self.ctx = Context(**kwds)
        
        self._surface = surface
        self._loop = loop
        
        
    def __str__(self):
        sP= str(self.surface.P)
        if len(sP) >40:
            sP = sP[:20] +"<...>"+ sP[-20:]
        s = "Elliptic surface with defining equation " + sP + " and a loop"
        return s
    
    def __repr__(self):
        sP= str(self.P)
        if len(sP) >40:
            sP = sP[:20] +"<...>"+ sP[-20:]
        s = "Elliptic surface with defining equation " + sP + " and a loop"
        return s
    
    @property
    def surface(self):
        return self._surface

    def unwinding_braid(self, loop):
        """Given a loop as a product of simple loops, determine the braid that puts this loop in simples form gamma1...gammak"""


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from ore_algebra import *
from sage.rings.complex_mpfr import ComplexField
from sage.functions.other import floor
from sage.arith.misc import gcd
from sage.arith.misc import xgcd
from sage.combinat.integer_vector import IntegerVectors
from sage.matrix.constructor import matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import identity_matrix
from sage.modules.free_module_element import vector
from sage.rings.integer_ring import ZZ

from sage.misc.prandom import randint, shuffle

from .numperiods.integerRelations import IntegerRelations

import logging

logger = logging.getLogger(__name__)


class Util(object):

    @classmethod 
    def simple_rational(cls, p, r):
        """Gives a rational q such that |p-q|<=r"""
        x=p
        l=[floor(x)]
        while abs(p-cls.evaluate_continued_fraction(l))>r:
            x=1/(x-l[-1])
            try:
                l+=[ZZ(x)]
            except:
                l+=[ZZ(floor(x))]
        return cls.evaluate_continued_fraction(l)
    
    @classmethod
    def evaluate_continued_fraction(cls, l):
        """ Given a list l, evaluates the continued fraction l[0] + 1/(l[1] + 1/(l[2] + ...))"""
        p=l[-1]
        l=l[:-1]
        while len(l)>0:
            p= l[-1] +1/p
            l=l[:-1]
        return p

    @classmethod
    def invert_permutation(cls, l):
        """Given a list representing a permutation of [0, ..., len(l)-1], l[i] = j, returns the inverse permutation l2[j] = i"""
        return [l.index(x) for x in range(len(l))]

    @classmethod
    def simplify_path(cls, p):
        """Given a list of numbers p, returns a ``simplification'' of p, removing all the backtracking
        
        For example, `simplify_path([1,2,3,2,1,4,5,1]) == [1,4,5,1]`
        """
        i=1
        res = list(p)
        while i<len(res)-1:
            p = res[i-1]
            a = res[i]
            n = res[i+1] 
            if p==a:
                res = res[:i]+res[i+1:]
                if i!=1:
                    i=i-1
            elif p==n:
                res = res[:i]+res[i+2:]
                if i!=1:
                    i=i-1
            else:
                i=i+1
        return res

    @classmethod
    def monomials(cls, ring, degree):
        """Given a polynomial ring and an integer d, returns all the monomials of the ring with degree d."""
        return [ring.monomial(*m) for m in list(IntegerVectors(degree, ring.ngens()))]

    @classmethod
    def xgcd_list(cls, l):
        """Given a list of integers l, return a double consisting of the gcd of these numbers and the coefficients of a Bezout relation."""
        if len(l)==0:
            return 0, []
        if len(l)==1:
            return l[0], [1]
        d = gcd(l)
        result = [1]
        a = l[0]
        for i in range(len(l)-1):
            b = l[i+1]
            d2, u, v = xgcd(a,b)
            result = [k*u for k in result] + [v]
            a=d2
        assert d2==d, "not getting the correct gcd"
        return d2, result


    @classmethod
    def path(cls, path, x):
        """Given a list of complex numbers path, and a number 0<x<1, return the path(x) where path is seen as a path [0,1]\\to C."""
        CC=ComplexField(500)
        dtot = sum([CC(abs(p1-p2)) for (p1,p2) in zip(path[:-1], path[1:])])
        dmin, dmax = 0, CC(abs(path[0]-path[1]))
        for i in range(len(path)-1):
            if x*dtot<=dmax and x*dtot>=dmin:
                break;
            else:
                dmin, dmax=dmax, dmax+CC(abs(path[i+1]-path[i+2]))
        t = Util.simple_rational((x*dtot -dmin)/(dmax-dmin), 10e-10)
        return (1-t)*path[i] + t*path[i+1]



    @classmethod
    def select_closest(cls, l, e):
        """Given a list of complex numbers l and a complex number e, return the element e2 of l minimizing abs(e2-e)"""
        # find element in l that is closest to e for abs
        CC=ComplexField(500)
        r = l[0]
        for i in range(1,len(l)):
            if abs(CC(l[i]-e))<abs(CC(r-e)):
                r = l[i]
        return r

    @classmethod
    def select_closest_index(cls, l, e):
        """Given a list of complex numbers l and a complex number e, return the index i minimizing abs(l[i]-e)"""
        # find index of element in l that is closest to e for abs
        CC=ComplexField(500)
        r = 0
        for i in range(1,len(l)):
            if abs(CC(l[i]-e))<abs(CC(l[r]-e)):
                r = i
        return r

    @classmethod
    def is_clockwise(cls, l):
        """Given a list of complex numbers describing a convex polygon, return whether the points are clockwise."""
        CC=ComplexField(500)
        smally = min(l, key=lambda v:(CC(v).imag(), CC(v).real()))
        i = l.index(smally)
        n = l[i+1 if i+1<len(l) else 0]
        p = l[i-1]

        x1,x2,x3 = [v.real() for v in [p,smally,n]]
        y1,y2,y3 = [v.imag() for v in [p,smally,n]]

        M = matrix([[1,x1,y1],[1,x2,y2],[1,x3,y3]])
        if abs(CC(M.determinant()))<10e-7:
            logger.info("cross product is very small, not certain about orientation")
        
        return CC(M.determinant())<0

    @classmethod
    def is_simple(cls, l):
        """Given a list of words l, return whether every word in the list consists of a single letter."""
        for w in l:
            if len(w.syllables()) != 1:
                return False
        return True
    
    @classmethod
    def letter(cls, w,i):
        """Given a word w and an integer i, yields the i-th letter of w."""
        return w.syllables()[i][0]**(w.syllables()[i][1]/abs(w.syllables()[i][1]))

    @classmethod
    def compatibility(cls, t1, t2, phi):
        w1, w2 = phi(t1), phi(t2)
        res = []
        if cls.letter(w1, 0) == cls.letter(w2, 0):
            res += [t2**-1*t1]
        if cls.letter(w1, -1) == cls.letter(w2, 0)**-1:
            res += [t1*t2]
        if cls.letter(w1, 0) == cls.letter(w2, -1)**-1:
            res += [t2*t1]
        if cls.letter(w1, -1) == cls.letter(w2, -1):
            res += [t2*t1**-1]
        return res
    
    @classmethod
    def easy_simplifications(cls, phi, ts=None):
        if ts == None:
            ts = list(phi.domain().gens())
        while not cls.is_simple([phi(t) for t in ts]):
            managed = False
            for i, t in enumerate(ts): 
                others = [t for j,t in enumerate(ts) if i != j]
                while len(phi(t).syllables())!=1:
                    options = [t*t2 for t2 in others if cls.letter(phi(t2),0) == cls.letter(phi(t),-1)**-1]
                    options += [t*t2**-1 for t2 in others if cls.letter(phi(t2**-1),0) == cls.letter(phi(t),-1)**-1]
                    options += [t2*t for t2 in others if cls.letter(phi(t2),-1) == cls.letter(phi(t),0)**-1]
                    options += [t2**-1*t for t2 in others if cls.letter(phi(t2**-1),-1) == cls.letter(phi(t),0)**-1]
                    if len(options)==0:
                        break
                    options = [o for o in options if phi(o)!=phi(1)]
                    options.sort(key = lambda w: len(phi(w).syllables()))

                    if len(phi(options[0]).syllables())< len(phi(t).syllables()):
                        t=options[0]
                        managed=True
                    else:
                        break
                ts[i] = t
            if not managed:
                break
        return ts

    @classmethod
    def lettersof(cls, w):
        letters = []
        for x, n in w.syllables():
            if x not in letters:
                letters += [x]
        return letters
    
    @classmethod
    def number_of_occurences(cls, w, x):
        res = 0
        for x2,n in w.syllables():
            if x2 == x:
                res += abs(n)
        return res
    
    @classmethod
    def invert_morphism(cls, phi):
        """Given an invertible free group morphism phi, computes its inverse. 
        Optionally, you can give generators ts of a subgroup (as a list of words) to compute the inverse of the restriction of phi on \\langle ts \\rangle (assuming it is invertible)."""
        singlets = []
        ts = cls.easy_simplifications(phi)
        xs = list(phi.codomain().gens())
        xs.sort(key = lambda x: len([t for t in ts if x in cls.lettersof(phi(t))]))
        replace = [1]
        while replace != []:
            done_something = False
            for x in xs:
                replace = []
                tsx = [t for t in ts if x in cls.lettersof(phi(t))]
                tsx.sort(key = lambda t: cls.number_of_occurences(phi(t), x))
                for t in tsx[1:]:
                    for newt in cls.compatibility(tsx[0], t, phi):
                        if cls.number_of_occurences(phi(newt), x) < cls.number_of_occurences(phi(t), x):
                            replace += [[t, newt]]
                            done_something = True
                            break
                if done_something:
                    break

            for i, t in enumerate(ts):
                for t2, newt in replace:
                    if t==t2:
                        ts[i]=newt
            for x in xs:
                tsx = [t for t in ts if x in cls.lettersof(phi(t))]
                if len(tsx) == 1:
                    singlets += [ts.pop(ts.index(tsx[0]))]
            xs.sort(key=lambda x: len([t for t in ts if x in cls.lettersof(phi(t))]))
        assert len(ts) == 0, "did not manage inversion"
        res = cls.easy_simplifications(phi, list(reversed(singlets)))

        tfin = [None]*len(res)
        for t in res:
            x, power = phi(t).syllables()[0]
            tfin[phi.codomain().gens().index(x)] = t**power
        return  phi.codomain().hom(tfin)


    @classmethod
    def find_complement(cls, B, primitive=True):
        """Given an m x n integer valued matrix B with n>m, computes an (n-m) x n matrix A such that the matrix block_matrix([[A],[B]]) is invertible over the integers"""
        D, U, V = B.smith_form()
        quotient = identity_matrix(D.ncols())[D.nrows():] * V.inverse()
        if primitive:
            assert block_matrix([[B],[quotient]]).det() in [-1,1], "cannot find complement, are you sure sublattice is primitive?"
        return quotient
    
    @classmethod
    def middle(self, w):
        """Given a word w of odd length 2n+1, yields the word consisting of the first n letters of w."""
        syls = w.syllables()
        syls = syls[:len(syls)//2]
        conj = w.parent(1)
        for l, p in syls:
            conj = conj*l**p
        return conj
    
    @classmethod
    def remove_duplicates(cls, l):
        l2 = []
        for e in l:
            if e not in l2:
                l2 += [e]
        return l2
    
    @classmethod
    def get_coefficient(cls, c, values, symbols):
        relations = IntegerRelations(matrix([c] + values).transpose()).basis
        assert relations.nrows()>=1, "could not identify coefficients"
        relation = relations[0]
        return -vector(relation[1:])*vector(symbols)/relation[0]
    
    @classmethod
    def rationalize(cls, c):
        return cls.get_coefficient(c,[1],[1])

    @classmethod
    def saturate(cls, Ms):
        span = identity_matrix(4).image()
        fam = identity_matrix(4).rows()
        new = True
        while new:
            new=False
            for v in span.span(fam).basis():
                for M in Ms:
                    if M*v not in span.span(fam):
                        new = True
                        fam += [M*v]
        CB = span.span(fam).basis_matrix().transpose()
        return CB
    
    @classmethod
    def check_if_algebraic(cls, c, order=10):
        try:
            return IntegerRelations(matrix([c**i for i in range(order+1)]).transpose()).basis.row(0)
        except:
            raise NotImplementedError("Non-algebraic number")


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from sage.modules.free_module_element import vector
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix
from sage.misc.misc_c import prod
from sage.misc.flatten import flatten


from .util import Util

import logging
import time

logger = logging.getLogger(__name__)


class MonodromyRepresentation(object):
    def __init__(self, monodromy_matrices, intersection_product):
        """`monodromy_matrices` is a list of matrices defining a monodromy representation
        """
        
        assert prod(list(reversed(monodromy_matrices)))==1, "monodromy representation is badly defined"
        self._monodromy_matrices = monodromy_matrices
        self._fibre_intersection_product = intersection_product
        self._dim = monodromy_matrices[0].nrows()
    
    
    @property
    def monodromy_matrices(self):
        return self._monodromy_matrices

    @property
    def fibre_intersection_product(self):
        return self._fibre_intersection_product
    
    @property
    def dim(self):
        return self._dim
    
    @property
    def homology(self): 
        if not hasattr(self, '_homology'):
            self._homology = identity_matrix(len(self.extensions_desingularisation) + self.add).rows()
        return self._homology

    def lift(self, v): 
        """Given a combination of thimbles of desingularisation, gives the corresponding homology class"""
        infinity_loops = matrix(self.infinity_loops).image().basis()
        infinity_loops = [self.desingularise(v) for v in infinity_loops]
        v = matrix(self.extensions_desingularisation + infinity_loops).solve_left(v)
        return vector(list(v)[:-len(infinity_loops)] + [0 for i in range(self.add)])

    @property
    def intersection_product(self):
        if not hasattr(self,'_intersection_product'):
            self._intersection_product = self._compute_intersection_product()
        return self._intersection_product
    
    @property
    def intersection_product_extensions(self):
        if not hasattr(self,'_intersection_product_extensions'):
            self._intersection_product_extensions = self._compute_intersection_product_extensions()
        return self._intersection_product_extensions

    @property
    def thimbles(self):
        if not hasattr(self,'_thimbles'):
            self._thimbles=[]
            for i, pcs in enumerate(self.permuting_cycles):
                for pc in pcs:
                    self._thimbles+=[(pc, i)]
        return self._thimbles
    
    @property
    def permuting_cycles(self):
        if not hasattr(self, '_permuting_cycles'):
            permuting_cycles = [[] for i in range(len(self.monodromy_matrices))]
            for i in range(len(self.monodromy_matrices)):
                M = self.monodromy_matrices[i]
                D, U, V = (M-1).smith_form()
                for j in range(self.dim):
                    if D[j,j]!=0:
                        permuting_cycles[i] += [ V * vector([1 if k==j else 0 for k in range(self.dim)]) ]
                permuting_cycles[i] = matrix(permuting_cycles[i]).image().gens()
            self._permuting_cycles = permuting_cycles
        return self._permuting_cycles

    @property
    def infinity_loops(self):
        """The linear combinations of thimbles that correspond to extensions along the (trivial) loop around infinity."""
        if not hasattr(self, '_infinity_loops'):
            infinity_cycles = []
            for i in range(self.dim):
                v = vector([1 if k==i else 0 for k in range(self.dim)])
                coefs = []
                for j in range(len(self.monodromy_matrices)):
                    M = self.monodromy_matrices[j]
                    if len(self.permuting_cycles[j])==0:
                        continue
                    coefs += list(matrix([ (M-1) * t for t in self.permuting_cycles[j]]).solve_left( (M-1) * v ))
                    v = self.monodromy_matrices[j]*v
                infinity_cycles+=[vector(coefs)]
            self._infinity_loops = matrix(infinity_cycles).change_ring(ZZ).rows()
        return self._infinity_loops
    
    @property
    def kernel_boundary(self):
        if not hasattr(self, '_kernel_boundary'):
            delta = matrix(self.vanishing_cycles).change_ring(ZZ)
            self._kernel_boundary = delta.kernel()
        return self._kernel_boundary
    

    @property
    def extensions(self):
        """Representants of the extensions of the elliptic surface."""
        if not hasattr(self, '_extensions'):
            delta = matrix(self.borders_of_thimbles).change_ring(ZZ)
            kerdelta = delta.kernel().matrix()
            D, U, V = kerdelta.smith_form()
            infinity_loops = matrix(self.infinity_loops).image().saturation().basis() # taking saturation here. This should in principle not be necessary
            B = D.solve_left(matrix(infinity_loops) * V).change_ring(ZZ)*U
            quotient_basis = Util.find_complement(B)
            if quotient_basis.nrows()==0:
                self._extensions = kerdelta.submatrix(0,0,0).rows()
            else:
                self._extensions = (quotient_basis*kerdelta).rows()
        return self._extensions
    
    @property
    def extensions_desingularisation(self):
        """Representant of the extensions of the desingularisation of the elliptic surface. 
        Along with the fibre and section, this constitutes a basis for the second homology group of the surface. 
        The singular fibre components are identified at the end of the list."""
        if not hasattr(self, '_extensions_desingularisation'):
            infinity_loops = matrix(self.infinity_loops).image().basis()
            infinity_loops = [self.desingularise(v) for v in infinity_loops]
            delta = matrix(flatten(self.vanishing_cycles_desingularisation)).change_ring(ZZ)
            kerdelta = delta.kernel().matrix()
            D, U, V = kerdelta.smith_form()
            B = D.solve_left(matrix(infinity_loops) * V).change_ring(ZZ)*U
            quotient_basis = Util.find_complement(B)
            if quotient_basis.nrows()==0:
                self._extensions_desingularisation = kerdelta.submatrix(0,0,0).rows()
            else:
                self._extensions_desingularisation = (quotient_basis*kerdelta).rows()
        return self._extensions_desingularisation


    def _compute_intersection_product_extensions(self):
        r=len(self.thimbles)
        extensions = matrix(self.extensions)
        inter_prod_thimbles = matrix([[self._compute_intersection_product_thimbles_extensions(i,j) for j in range(r)] for i in range(r)])
        intersection_11 = (extensions * inter_prod_thimbles * extensions.transpose()).change_ring(ZZ)
        return intersection_11

    def _compute_intersection_product_thimbles_extensions(self, i, j):
        vi, loopi = self.thimbles[i]
        vj, loopj = self.thimbles[j]

        Mi = self.monodromy_matrices[loopi]
        Mj = self.monodromy_matrices[loopj]

        di, dj = (Mi-1)*vi, (Mj-1)*vj

        res = di * self.fibre_intersection_product * dj
        resid = -vi * self.fibre_intersection_product * dj

        if loopi == loopj:
            return resid
        if loopi < loopj:
            return res
        else:
            return 0


    def _compute_intersection_product(self):
        r=len(flatten(self.vanishing_cycles_desingularisation))
        extensions = matrix(self.extensions_desingularisation)
        inter_prod_thimbles = matrix([[self._compute_intersection_product_thimbles(i,j) for j in range(r)] for i in range(r)])
        intersection_11 = (-1 if self.add==2 else 1) * (extensions * inter_prod_thimbles * extensions.transpose()).change_ring(ZZ)
        if self.add==2:
            intersection_02 = zero_matrix(2,2)
            intersection_02[0,1], intersection_02[1,0] = 1,1
            intersection_02[1,1] = self.self_intersection_section
            return block_diagonal_matrix(intersection_11, intersection_02)
        return intersection_11
        
    def _compute_intersection_product_thimbles(self,i,j):
        vi = self.permuting_cycles_desingularisation[i]
        Mi = flatten(self.monodromy_matrices_desingularisation)[i]
        vj = self.permuting_cycles_desingularisation[j]
        Mj = flatten(self.monodromy_matrices_desingularisation)[j]
        di, dj = (Mi-1) * vi, (Mj-1) * vj
        res = di*self.fibre_intersection_product*dj
        resid = -vi*self.fibre_intersection_product*di

        if i==j:
            return resid
        if i<j:
            return res
        else:
            return 0

    
    @property
    def primary_lattice(self):
        if not hasattr(self, '_primary_lattice'):
            extensions = [self.lift(self.desingularise(v)) for v in self.extensions]
            components_of_singular_fibres = [self.lift(v) for v in flatten(self.components_of_singular_fibres, max_level=2)]
            primary_lattice = extensions + components_of_singular_fibres
            if self.add==2:
                primary_lattice += [self.fibre_class, self.section]
            self._primary_lattice = matrix(primary_lattice).transpose()
        return self._primary_lattice

    def desingularise(self, v):
        """Given an extension of the surface, yields its description as an extension of the desingularisation."""
        return v * self.thimbles_confluence
    
    @property
    def permuting_cycles_desingularisation(self):
        if not hasattr(self, '_permuting_cycles_desingularisation'):
            monodromy_matrices = flatten(self.monodromy_matrices_desingularisation)
            vanishing = flatten(self.vanishing_cycles_desingularisation)
            _permuting_cycles_desingularisation = []
            for i in range(len(monodromy_matrices)):
                M = monodromy_matrices[i]
                D, U, V = (M-1).smith_form()
                p = V.column(0)
                if (M-1) * p != vanishing[i]:
                    p = -p
                assert (M-1) * p == vanishing[i]
                _permuting_cycles_desingularisation += [ p ]
            self._permuting_cycles_desingularisation = _permuting_cycles_desingularisation
        return self._permuting_cycles_desingularisation

    def lift(self, v):
            """Given a combination of thimbles of desingularisation, gives the corresponding homology class"""
            infinity_loops = matrix(self.infinity_loops).image().basis()
            infinity_loops = [self.desingularise(v) for v in infinity_loops]
            v = matrix(self.extensions_desingularisation + infinity_loops).solve_left(v)
            add = [0,0] if self.add==2 else []
            return vector(list(v)[:-len(infinity_loops)] + add)

    @property
    def thimbles_confluence(self):
        if not hasattr(self, '_thimbles_confluence'):
            blocks =[]
            for i, pcs in enumerate(self.permuting_cycles):
                decompositions = []
                for p in pcs:
                    decomposition = []
                    for M, v in zip(self.monodromy_matrices_desingularisation[i], self.vanishing_cycles_desingularisation[i]):
                        decomposition += [(M-1) * p / v]
                        p = M * p
                    decompositions+=[decomposition]
                blocks+= [matrix(decompositions)]
            self._thimbles_confluence = block_diagonal_matrix(blocks).change_ring(ZZ)
        return self._thimbles_confluence

    @property
    def vanishing_cycles_desingularisation(self):
        if not hasattr(self, '_vanishing_cycles_desingularisation'):
            self._vanishing_cycles_desingularisation = [[(M-1).transpose().image().gens()[0] for M in Ms] for Ms in self.monodromy_matrices_desingularisation]
        return self._vanishing_cycles_desingularisation
    
    @property
    def borders_of_thimbles(self):
        if not hasattr(self, '_borders_of_thimbles'):
            self._borders_of_thimbles = []
            for ps, M in zip(self.permuting_cycles, self.monodromy_matrices):
                self._borders_of_thimbles += [ (M-1) * p for p in ps]
        return self._borders_of_thimbles

    @property
    def components_of_singular_fibres(self):
        if not hasattr(self, '_components_of_singular_fibres'):
            ranktot = 0
            rankmax = sum([len(Ms) for Ms in self.monodromy_matrices_desingularisation])
            sing_comps = []
            for M in self.vanishing_cycles_desingularisation:
                M = matrix(M)
                rank = M.dimensions()[0]
                sing_comps += [[vector([0]*ranktot+list(v) + [0]*(rankmax-ranktot-rank)) for v in M.kernel().gens()]]
                ranktot+=rank
            self._components_of_singular_fibres = sing_comps
        return self._components_of_singular_fibres
    
    @property
    def fibre_class(self):
        assert self.add == 2, "no fibre class in odd dimensions"
        return vector([0] * len(self.extensions + flatten(self.components_of_singular_fibres)) + [1,0])
    
    @property
    def section(self):
        assert self.add == 2, "no section in odd dimensions"
        return vector([0] * len(self.extensions + flatten(self.components_of_singular_fibres)) + [0,1])
    
    @property
    def monodromy_matrices_desingularisation(self):
        if not hasattr(self, '_monodromy_matrices_desingularisation'):
            monodromy_matrices_desingularisation = []
            for M in self.monodromy_matrices:
                decomposition = self.desingularise_matrix(M)
                assert prod(list(reversed(decomposition))) == M
                monodromy_matrices_desingularisation += [decomposition]
            self._monodromy_matrices_desingularisation = monodromy_matrices_desingularisation
        return self._monodromy_matrices_desingularisation

    def desingularise_matrix(self, M):
        raise NotImplementedError("`desingularise_matrix` not implemented")